
'''
def moveTo(self, point):
    """
    Move to a point `x`, `y`.
    """
    super(BezierPath, self).moveTo(point)

def lineTo(self, point):
    """
    Line to a point `x`, `y`.
    """
    super(BezierPath, self).lineTo(point)

def curveTo(self, *points):
    """
    Draw a cubic Bézier with an arbitrary number of control points.

    The last point specified is on-curve, all others are off-curve
    (control) points.
    """
    super(BezierPath, self).curveTo(*points)

def qCurveTo(self, *points):
    """
    Draw a whole string of quadratic curve segments.

    The last point specified is on-curve, all others are off-curve
    (control) points.
    """
    super(BezierPath, self).qCurveTo(*points)

def addComponent(self, glyphName, transformation):
    """
    Add a sub glyph. The 'transformation' argument must be a 6-tuple
    containing an affine transformation, or a Transform object from the
    fontTools.misc.transform module. More precisely: it should be a
    sequence containing 6 numbers.

    A `glyphSet` is required during initialization of the BezierPath object.
    """
    super(BezierPath, self).addComponent(glyphName, transformation)
'''



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# -----------------------------------------------------------------------------
#
#     P A G E B O T
#
#     Copyright (c) 2016+ Buro Petr van Blokland + Claudia Mens
#     www.pagebot.io
#     Licensed under MIT conditions
#
#     Supporting DrawBot, www.drawbot.com
#     Supporting Flat, xxyxyz.org/flat
# -----------------------------------------------------------------------------
#
#     osxbezierpath.py
#

import math
import AppKit
import Quartz
import CoreText
import booleanOperations
from fontTools.pens.basePen import BasePen
from fontTools.pens.pointPen import PointToSegmentPen
from pagebot.contexts.base.basecontext import BaseContext
from pagebot.contexts.base.basepoint import BasePoint
from drawBot.context.tools import traceImage
from pagebotosx.errors import PageBotOSXError
from pagebotosx.bezierpaths.beziercontour import BezierContour
from pagebotosx.mathematics.transform import *

_FALLBACKFONT = "LucidaGrande"

class OSXBezierPath(BasePen):
    """A Bézier path object, if you want to draw the same over and over
    again."""

    contourClass = BezierContour

    _textAlignMap = dict(
        center=AppKit.NSCenterTextAlignment,
        left=AppKit.NSLeftTextAlignment,
        right=AppKit.NSRightTextAlignment,
        justified=AppKit.NSJustifiedTextAlignment,
    )

    _instructionSegmentTypeMap = {
        AppKit.NSMoveToBezierPathElement: "move",
        AppKit.NSLineToBezierPathElement: "line",
        AppKit.NSCurveToBezierPathElement: "curve"
    }

    def __init__(self, path=None, glyphSet=None):
        if path is None:
            self._path = AppKit.NSBezierPath.alloc().init()
        else:
            self._path = path
        BasePen.__init__(self, glyphSet)

    def __repr__(self):
        return "<OSXBezierPath>"

    # pen support.

    def _moveTo(self, pt):
        self._path.moveToPoint_(pt)

    def _lineTo(self, pt):
        self._path.lineToPoint_(pt)

    def _curveToOne(self, pt1, pt2, pt3):
        """Curve to a point `x3`, `y3`. With given Bézier handles `x1`, `y1`
        and `x2`, `y2`."""
        self._path.curveToPoint_controlPoint1_controlPoint2_(pt3, pt1, pt2)

    # TODO: QCurve?

    def closePath(self):
        """Close the path."""
        self._path.closePath()

    def beginPath(self, identifier=None):
        """Begin using the path as a so called point pen and start a new
        subpath."""
        self._pointToSegmentPen = PointToSegmentPen(self)
        self._pointToSegmentPen.beginPath()

    def addPoint(self, point, segmentType=None, smooth=False, name=None,
            identifier=None, **kwargs):
        """Use the path as a point pen and add a point to the current subpath.
        `beginPath` must have been called prior to adding points with
        `addPoint` calls."""
        if not hasattr(self, "_pointToSegmentPen"):
            raise PageBotOSXError("path.beginPath() must be called before the path can be used as a point pen")
        self._pointToSegmentPen.addPoint(
            point,
            segmentType=segmentType,
            smooth=smooth,
            name=name,
            identifier=identifier,
            **kwargs
        )

    def endPath(self):
        """End the current subpath. Calling this method has two distinct
        meanings depending on the context:

        When the Bézier path is used as a segment pen (using `moveTo`,
        `lineTo`, etc.), the current subpath will be finished as an open
        contour.

        When the Bézier path is used as a point pen (using `beginPath`,
        `addPoint` and `endPath`), the path will process all the points added
        with `addPoint`, finishing the current subpath."""
        if hasattr(self, "_pointToSegmentPen"):
            # its been uses in a point pen world
            pointToSegmentPen = self._pointToSegmentPen
            del self._pointToSegmentPen
            pointToSegmentPen.endPath()
        else:
            # with NSBezierPath, nothing special needs to be done for an open subpath.
            pass

    def drawToPen(self, pen):
        """
        Draw the Bézier path into a pen
        """
        contours = self.contours
        for contour in contours:
            contour.drawToPen(pen)

    def drawToPointPen(self, pointPen):
        """Draw the Bézier path into a point pen."""
        contours = self.contours
        for contour in contours:
            contour.drawToPointPen(pointPen)

    def arc(self, center, radius, startAngle, endAngle, clockwise):
        """Arc with `center` and a given `radius`, from `startAngle` to
        `endAngle`, going clockwise if `clockwise` is True and counter
        clockwise if `clockwise` is False."""
        self._path.appendBezierPathWithArcWithCenter_radius_startAngle_endAngle_clockwise_(
            center, radius, startAngle, endAngle, clockwise)

    def arcTo(self, point1, point2, radius):
        """Arc defined by a circle inscribed inside the angle specified by
        three points: the current point, `point1`, and `point2`. The arc is
        drawn between the two points of the circle that are tangent to the two
        legs of the angle. """
        self._path.appendBezierPathWithArcFromPoint_toPoint_radius_(point1, point2, radius)

    def rect(self, x, y, w, h):
        """Add a rectangle at possition `x`, `y` with a size of `w`, `h`."""
        self._path.appendBezierPathWithRect_(((x, y), (w, h)))

    def oval(self, x, y, w, h):
        """Add a oval at possition `x`, `y` with a size of `w`, `h`."""
        self._path.appendBezierPathWithOvalInRect_(((x, y), (w, h)))
        self.closePath()

    def text(self, txt, offset=None, font=_FALLBACKFONT, fontSize=10, align=None):
        """Draws a `txt` with a `font` and `fontSize` at an `offset` in the
        Bézier path. If a font path is given the font will be installed and
        used directly.

        - Optionally an alignment can be set.
        - Possible `align` values are: `"left"`, `"center"` and `"right"`.
        - The default alignment is `left`.
        - Optionally `txt` can be a `FormattedString`.
        """
        context = BaseContext()

        if align and align not in self._textAlignMap.keys():
            raise PageBotOSXError("align must be %s" % (", ".join(self._textAlignMap.keys())))

        context.font(font, fontSize)
        attributedString = context.attributedString(txt, align)
        w, h = attributedString.size()
        w *= 2

        if offset:
            x, y = offset
        else:
            x = y = 0
        if align == "right":
            x -= w
        elif align == "center":
            x -= w * .5

        setter = CoreText.CTFramesetterCreateWithAttributedString(attributedString)
        path = Quartz.CGPathCreateMutable()
        Quartz.CGPathAddRect(path, None, Quartz.CGRectMake(x, y, w, h * 2))
        frame = CoreText.CTFramesetterCreateFrame(setter, (0, 0), path, None)
        ctLines = CoreText.CTFrameGetLines(frame)
        origins = CoreText.CTFrameGetLineOrigins(frame, (0, len(ctLines)), None)

        if origins:
            y -= origins[0][1]

        self.textBox(txt, box=(x, y, w, h * 2), font=font, fontSize=fontSize, align=align)

    def textBox(self, txt, box, font=_FALLBACKFONT, fontSize=10, align=None,
            hyphenation=None):
        """Draws a `txt` with a `font` and `fontSize` in a `box` in the Bézier
        path. If a font path is given the font will be installed and used
        directly.

        - Optionally an alignment can be set.
        - Possible `align` values are: `"left"`, `"center"` and `"right"`.
        - The default alignment is `left`.
        - Optionally `hyphenation` can be provided.
        - Optionally `txt` can be a `FormattedString`.
        - Optionally `box` can be a `BezierPath`.
        """
        if align and align not in self._textAlignMap.keys():
            raise PageBotOSXError("align must be %s" % (", ".join(self._textAlignMap.keys())))

        context = BaseContext()
        context.font(font, fontSize)
        context.hyphenation(hyphenation)
        path, (x, y) = context._getPathForFrameSetter(box)
        attributedString = context.attributedString(txt, align)

        setter = CoreText.CTFramesetterCreateWithAttributedString(attributedString)
        frame = CoreText.CTFramesetterCreateFrame(setter, (0, 0), path, None)
        ctLines = CoreText.CTFrameGetLines(frame)
        origins = CoreText.CTFrameGetLineOrigins(frame, (0, len(ctLines)), None)

        for i, (originX, originY) in enumerate(origins):
            ctLine = ctLines[i]
            ctRuns = CoreText.CTLineGetGlyphRuns(ctLine)
            for ctRun in ctRuns:
                attributes = CoreText.CTRunGetAttributes(ctRun)
                font = attributes.get(AppKit.NSFontAttributeName)
                baselineShift = attributes.get(AppKit.NSBaselineOffsetAttributeName, 0)
                glyphCount = CoreText.CTRunGetGlyphCount(ctRun)
                for i in range(glyphCount):
                    glyph = CoreText.CTRunGetGlyphs(ctRun, (i, 1), None)[0]
                    ax, ay = CoreText.CTRunGetPositions(ctRun, (i, 1), None)[0]
                    if glyph:
                        self._path.moveToPoint_((x + originX + ax, y + originY + ay + baselineShift))
                        self._path.appendBezierPathWithGlyph_inFont_(glyph, font)
        self.optimizePath()
        return context.clippedText(txt, box, align)

    def traceImage(self, path, threshold=.2, blur=None, invert=False, turd=2,
            tolerance=0.2, offset=None):
        """Converts a given image to a vector outline. Some tracing options can
        be provide:

        * `threshold`: the threshold used to bitmap an image
        * `blur`: the image can be blurred
        * `invert`: invert to the image
        * `turd`: the size of small turd that can be ignored
        * `tolerance`: the precision tolerance of the vector outline
        * `offset`: add the traced vector outline with an offset to the BezierPath
        """
        traceImage.TraceImage(path, self, threshold, blur, invert, turd, tolerance, offset)

    def getNSBezierPath(self):
        """
        Return the nsBezierPath.
        """
        return self._path

    def _getCGPath(self):
        path = Quartz.CGPathCreateMutable()
        count = self._path.elementCount()
        for i in range(count):
            instruction, points = self._path.elementAtIndex_associatedPoints_(i)
            if instruction == AppKit.NSMoveToBezierPathElement:
                Quartz.CGPathMoveToPoint(path, None, points[0].x, points[0].y)
            elif instruction == AppKit.NSLineToBezierPathElement:
                Quartz.CGPathAddLineToPoint(path, None, points[0].x, points[0].y)
            elif instruction == AppKit.NSCurveToBezierPathElement:
                Quartz.CGPathAddCurveToPoint(
                    path, None,
                    points[0].x, points[0].y,
                    points[1].x, points[1].y,
                    points[2].x, points[2].y
                )
            elif instruction == AppKit.NSClosePathBezierPathElement:
                Quartz.CGPathCloseSubpath(path)

        # hacking to get a proper close path at the end of the path
        x, y, _, _ = self.bounds()
        Quartz.CGPathMoveToPoint(path, None, x, y)
        Quartz.CGPathAddLineToPoint(path, None, x, y)
        Quartz.CGPathAddLineToPoint(path, None, x, y)
        Quartz.CGPathAddLineToPoint(path, None, x, y)
        Quartz.CGPathCloseSubpath(path)
        return path

    def setNSBezierPath(self, path):
        """Sets a nsBezierPath."""
        self._path = path

    def pointInside(self, xy):
        """Checks if a point `x`, `y` is inside a path."""
        x, y = xy
        return self._path.containsPoint_((x, y))

    def bounds(self):
        """Returns the bounding box of the path."""
        if self._path.isEmpty():
            return None
        (x, y), (w, h) = self._path.bounds()
        return x, y, x + w, y + h

    def controlPointBounds(self):
        """Returns the bounding box of the path including the offcurve
        points."""
        (x, y), (w, h) = self._path.controlPointBounds()
        return x, y, x + w, y + h

    def optimizePath(self):
        count = self._path.elementCount()
        if self._path.elementAtIndex_(count - 1) == AppKit.NSMoveToBezierPathElement:
            optimizedPath = AppKit.NSBezierPath.alloc().init()
            for i in range(count - 1):
                instruction, points = self._path.elementAtIndex_associatedPoints_(i)
                if instruction == AppKit.NSMoveToBezierPathElement:
                    optimizedPath.moveToPoint_(*points)
                elif instruction == AppKit.NSLineToBezierPathElement:
                    optimizedPath.lineToPoint_(*points)
                elif instruction == AppKit.NSCurveToBezierPathElement:
                    p1, p2, p3 = points
                    optimizedPath.curveToPoint_controlPoint1_controlPoint2_(p3, p1, p2)
                elif instruction == AppKit.NSClosePathBezierPathElement:
                    optimizedPath.closePath()
            self._path = optimizedPath

    def copy(self):
        """Copies the Bézier path."""
        new = self.__class__()
        new._path = self._path.copy()
        return new

    def reverse(self):
        """Reverses the path direction."""
        self._path = self._path.bezierPathByReversingPath()

    def appendPath(self, otherPath):
        """Appends a path."""
        self._path.appendBezierPath_(otherPath.getNSBezierPath())

    def __add__(self, otherPath):
        new = self.copy()
        new.appendPath(otherPath)
        return new

    def __iadd__(self, other):
        self.appendPath(other)
        return self

    # transformations

    def translate(self, x=0, y=0):
        """Translates the path with a given offset."""
        self.transform((1, 0, 0, 1, x, y))

    def rotate(self, angle, center=(0, 0)):
        """Rotates the path around the `center` point (which is the origin by
        default) with a given angle in degrees."""
        angle = math.radians(angle)
        c = math.cos(angle)
        s = math.sin(angle)
        self.transform((c, s, -s, c, 0, 0), center)

    def scale(self, x=1, y=None, center=(0, 0)):
        """Scales the path with a given `x` (horizontal scale) and `y` (vertical
        scale). If only 1 argument is provided a proportional scale is
        applied. The center of scaling can optionally be set via the `center`
        keyword argument. By default this is the origin."""
        if y is None:
            y = x
        self.transform((x, 0, 0, y, 0, 0), center)

    def skew(self, angle1, angle2=0, center=(0, 0)):
        """Skews the path with given `angle1` and `angle2`. If only one
        argument is provided a proportional skew is applied. The center of
        skewing can optionally be set via the `center` keyword argument. By
        default this is the origin."""
        angle1 = math.radians(angle1)
        angle2 = math.radians(angle2)
        self.transform((1, math.tan(angle2), math.tan(angle1), 1, 0, 0), center)

    def transform(self, transformMatrix, center=(0, 0)):
        """Transforms a path with a transform matrix (xy, xx, yy, yx, x, y)."""
        if center != (0, 0):
            transformMatrix = transformationAtCenter(transformMatrix, center)

        aT = AppKit.NSAffineTransform.alloc().init()
        aT.setTransformStruct_(transformMatrix[:])
        self._path.transformUsingAffineTransform_(aT)

    # boolean operations

    def _contoursForBooleanOperations(self):
        # contours are very temporaly objects
        # redirect drawToPointPen to drawPoints
        contours = self.contours
        for contour in contours:
            contour.drawPoints = contour.drawToPointPen
            if contour.open:
                raise PageBotOSXError("open contours are not supported during boolean operations")
        return contours

    def union(self, other):
        """Returns the union between two Bézier paths."""
        assert isinstance(other, self.__class__)
        contours = self._contoursForBooleanOperations() + other._contoursForBooleanOperations()
        result = self.__class__()
        booleanOperations.union(contours, result)
        return result

    def removeOverlap(self):
        """Removes all overlaps in a Bézier path."""
        contours = self._contoursForBooleanOperations()
        result = self.__class__()
        booleanOperations.union(contours, result)
        self.setNSBezierPath(result.getNSBezierPath())
        return self

    def difference(self, other):
        """Returns the difference between two Bézier paths."""
        assert isinstance(other, self.__class__)
        subjectContours = self._contoursForBooleanOperations()
        clipContours = other._contoursForBooleanOperations()
        result = self.__class__()
        booleanOperations.difference(subjectContours, clipContours, result)
        return result

    def intersection(self, other):
        """Returns the intersection between two Bézier paths."""
        assert isinstance(other, self.__class__)
        subjectContours = self._contoursForBooleanOperations()
        clipContours = other._contoursForBooleanOperations()
        result = self.__class__()
        booleanOperations.intersection(subjectContours, clipContours, result)
        return result

    def xor(self, other):
        """Returns the xor between two Bézier paths."""
        assert isinstance(other, self.__class__)
        subjectContours = self._contoursForBooleanOperations()
        clipContours = other._contoursForBooleanOperations()
        result = self.__class__()
        booleanOperations.xor(subjectContours, clipContours, result)
        return result

    def intersectionPoints(self, other=None):
        """Returns a list of intersection points as `x`, `y` tuples. Optionaly
        provides an other path object to find intersection points."""
        contours = self._contoursForBooleanOperations()
        if other is not None:
            assert isinstance(other, self.__class__)
            contours += other._contoursForBooleanOperations()
        return booleanOperations.getIntersections(contours)

    def __mod__(self, other):
        return self.difference(other)

    __rmod__ = __mod__

    def __imod__(self, other):
        result = self.difference(other)
        self.setNSBezierPath(result.getNSBezierPath())
        return self

    def __or__(self, other):
        return self.union(other)

    __ror__ = __or__

    def __ior__(self, other):
        result = self.union(other)
        self.setNSBezierPath(result.getNSBezierPath())
        return self

    def __and__(self, other):
        return self.intersection(other)

    __rand__ = __and__

    def __iand__(self, other):
        result = self.intersection(other)
        self.setNSBezierPath(result.getNSBezierPath())
        return self

    def __xor__(self, other):
        return self.xor(other)

    __rxor__ = __xor__

    def __ixor__(self, other):
        result = self.xor(other)
        self.setNSBezierPath(result.getNSBezierPath())
        return self

    def _points(self, onCurve=True, offCurve=True):
        points = []

        if not onCurve and not offCurve:
            return points

        for index in range(self._path.elementCount()):
            instruction, pts = self._path.elementAtIndex_associatedPoints_(index)
            if not onCurve:
                pts = pts[:-1]
            elif not offCurve:
                pts = pts[-1:]
            points.extend([BasePoint(p.x, p.y, onCurve=onCurve) for p in pts])
        return points

    def _get_points(self):
        return self._points()

    points = property(_get_points, doc="Return a list of all points.")

    def _get_onCurvePoints(self):
        return self._points(offCurve=False)

    onCurvePoints = property(_get_onCurvePoints, doc="Return a list of all on curve points.")

    def _get_offCurvePoints(self):
        return self._points(onCurve=False)

    offCurvePoints = property(_get_offCurvePoints, doc="Return a list of all off curve points.")

    def _get_contours(self):
        contours = []
        for index in range(self._path.elementCount()):
            instruction, pts = self._path.elementAtIndex_associatedPoints_(index)
            if instruction == AppKit.NSMoveToBezierPathElement:
                contours.append(self.contourClass())
            if instruction == AppKit.NSClosePathBezierPathElement:
                contours[-1].open = False
            if pts:
                contours[-1].append([(p.x, p.y) for p in pts])
        if len(contours) >= 2 and len(contours[-1]) == 1 and contours[-1][0] == contours[-2][0]:
            contours.pop()
        return contours

    contours = property(_get_contours, doc="Return a list of contours with all point coordinates sorted in segments. A contour object has an `open` attribute.")

    def __len__(self):
        return len(self.contours)

    def __getitem__(self, index):
        return self.contours[index]

    def __iter__(self):
        contours = self.contours
        count = len(contours)
        index = 0
        while index < count:
            contour = contours[index]
            yield contour
            index += 1



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# -----------------------------------------------------------------------------
#
#     P A G E B O T
#
#     Copyright (c) 2016+ Buro Petr van Blokland + Claudia Mens
#     www.pagebot.io
#     Licensed under MIT conditions
#
#     Supporting DrawBot, www.drawbot.com
#     Supporting Flat, xxyxyz.org/flat
# -----------------------------------------------------------------------------
#
#     svg2drawbot.py
#
#     Converts SVG paths to drawbot Bézier paths.
#
from xml.dom import minidom
from pagebotosx.bezierpaths.osxbezierpath import OSXBezierPath

def getSvgPaths(fileName):
    """Extracts path strings from XML."""
    doc = minidom.parse(fileName)  # parseString also exists
    svgPaths = [path.getAttribute('d') for path
                in doc.getElementsByTagName('path')]
    doc.unlink()
    return svgPaths

def parseSVG(strings):
    """Takes a list of path strings and converts them to a list of SVG-command
    tuples."""
    cmd =['m', 'l', 'v', 'c', 'h', 'z', 's']

    paths = []

    for string in strings:
        command = None   # Current command.
        valuestring = ''
        path = []
        points = []

        for c in string:

            if c.lower() in cmd:
                # New command, add previous one to path.
                if  command is not None:
                    path.append((command, points))

                command = c
                addValueToPoints(valuestring, points)

                # Reset.
                points = []
                valuestring = ''
            else:
                # Skip spaces.
                if c == ' ':
                    continue

                # New value.
                if c in (',', '-'):
                    addValueToPoints(valuestring, points)

                    # Split on minus.
                    if c == '-':
                        valuestring = c
                    else:
                        valuestring = ''
                else:
                    valuestring += c

        paths.append(path)
    return paths

def reflect(point0, point1):
    """Reflects off-curve control point in relation to on-curve one. Used for
    smooth curves."""
    px = point1[0] + (point1[0] - point0[0])
    py = point1[1] + (point1[1] - point0[1])
    return (px, py)

def getRelative(points, pPrevious):
    """Calculates absolute coordinates by adding the previous point to the
    relative coordinates."""
    newPoints = []

    for p in points:
        newP = (p[0] + pPrevious[0], p[1] + pPrevious[1])
        newPoints.append(newP)

    return newPoints

def copyPoint(previousPoint, currentPoint):
    """Deep copies previous point."""
    previousPoint[0] = currentPoint[0]
    previousPoint[1] = currentPoint[1]

def contourToPath(contour):
    """Converts SVG contour to a path in DrawBot."""
    path = OSXBezierPath()
    pPrev = [0.0, 0.0]
    pPrev2 = None
    previousCommand = None

    for segment in contour:
        command = segment[0]
        points = segment[1]

        relative = False

        if command.islower():
            relative = True

        command = command.lower()

        if command == 'm':
            if relative:
                points = getRelative(points, pPrev)

            path.moveTo(points[0])
        elif command in ('l', 'h', 'v'):
            if command == 'l':
                if relative:
                    points = getRelative(points, pPrev)
            elif command == 'h':
                if relative:
                    points[0][0] += pPrev[0]

                points[0].append(pPrev[1])
            elif command == 'v':
                points[0].insert(0, pPrev[0])

                if relative:
                    points[0][1] += pPrev[1]

            path.lineTo(points[0])
        elif command == 'c':
            if relative:
                points = getRelative(points, pPrev)

            path.curveTo(*points)
        elif command == 's':
            if relative:
                points = getRelative(points, pPrev)

            # TODO: From the spec:
            #
            # If there is no previous command or if the previous command was
            # not an C, c, S or s, assume the first control point is coincident
            # with the current point
            if previousCommand in ('c', 's'):
                cp = reflect(pPrev2, pPrev)
            else:
                print('TODO: implement, test')
                #copyPoint(cp, points[0])
            path.curveTo(cp, *points)

        elif command == 'z':
            continue

        copyPoint(pPrev, points[-1])

        if command in ('c', 's'):
            pPrev2 = [0.0, 0.0]
            copyPoint(pPrev2, points[-2])
        else:
            pPrev2 = None

        previousCommand = command

    path.closePath()
    return path

def addValueToPoints(valuestring, points):
    """Adds the collected character string to the last coordinate
    in the points list."""
    if not valuestring:
        return

    value = float(valuestring)

    if not points or len(points[-1]) == 2:
        points.append([])

    points[-1].append(value)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# -----------------------------------------------------------------------------
#
#     P A G E B O T
#
#     Copyright (c) 2016+ Buro Petr van Blokland + Claudia Mens
#     www.pagebot.io
#     Licensed under MIT conditions
#
#     Supporting DrawBot, www.drawbot.com
#     Supporting Flat, xxyxyz.org/flat
# -----------------------------------------------------------------------------
#
#     textrun.py
#

from CoreText import (CTRunGetGlyphCount, CTRunGetStringRange,
        CTRunGetStringIndicesPtr, CTRunGetAdvances, CTRunGetStatus,
        CTRunGetPositions, CTRunGetGlyphs, CTRunGetAttributes,
        CTRunGetTextMatrix, CFRange)
from pagebot.toolbox.units import em, pt
from pagebot.toolbox.color import color
from pagebot.fonttoolbox.objects.font import findFont

class TextRun:
    """A TextRun object contains the chunk of text that combines a single
    style.

    Wraps the CoreText CTRun class, see also:

    https://developer.apple.com/documentation/coretext/ctrun-61n
    """

    def __init__(self, ctRun, runIndex):
        # Index of the run in the TextLine.
        self.runIndex = runIndex
        self._ctRun = ctRun

        # Property cash for constructed style from run parameters.
        self._style = None
        self.glyphCount = gc = CTRunGetGlyphCount(ctRun)

        # Reverse the style from
        attrs = CTRunGetAttributes(ctRun)
        self.nsFont = attrs['NSFont']

        # Initialize by calling the property self.font.
        self._font = None

        #self.fontDescriptor = f.fontDescriptor()
        self.nsColor = attrs['NSColor']
        # Initialize by calling self.color.
        self._fill = None
        self.nsParagraphStyle = attrs['NSParagraphStyle']
        # Save, in case the caller want to query other run parameters.
        self.attrs = attrs

        self.iStart, self.iEnd = CTRunGetStringRange(ctRun)
        self.string = ''

        # Hack for now to find the string in repr-string if self._ctLine.
        # TODO: Make a better conversion here, not relying on the format of the
        # repr-string.
        for index, part in enumerate(str(ctRun).split('"')[1].split('\\u')):
            if index == 0:
                self.string += part
            elif len(part) >= 4:
                self.string += chr(int(part[0:4], 16))
                self.string += part[4:]

        #print(gc, len(CTRunGetStringIndicesPtr(ctRun)), CTRunGetStringIndicesPtr(ctRun), ctRun)
        try:
            self.stringIndices = CTRunGetStringIndicesPtr(ctRun)[0:gc]
        except TypeError:
            self.stringIndices = [0]

        #CTRunGetStringIndices(ctRun._ctRun, CFRange(0, 5), None)[4]
        self.advances = CTRunGetAdvances(ctRun, CFRange(0, 5), None)
        #self.positions = CTRunGetPositionsPtr(ctRun)[0:gc]
        #CTRunGetPositions(ctRun, CFRange(0, 5), None)[4]
        #self.glyphFontIndices = CTRunGetGlyphsPtr(ctRun)[0:gc]
        #print(CTRunGetGlyphs(ctRun, CFRange(0, 5), None)[0:5])
        self.status = CTRunGetStatus(ctRun)

        # get all positions
        self.positions = CTRunGetPositions(ctRun, (0, gc), None)
        # get all glyphs
        self.glyphs = CTRunGetGlyphs(ctRun, (0, gc), None)

    def __len__(self):
        return self.glyphCount

    def __repr__(self):
        return '[TextRun #%d "%s"]' % (self.runIndex, self.string)

    def __getitem__(self, index):
        return self.string[index]

    def _get_style(self):
        """Answers the constructed style dictionary, with names that fit the
        standard PageBot style."""
        if self._style is None:
            self._style = dict(
                textFill=self.textFill,
                pl=self.headIndent,
                pr=self.tailIndent,
                fontSize=self.fontSize,
                font=self.font or self.fontName, # Answer Font object or font name.
                leading=self.leading
            )
        return self._style
    style = property(_get_style)

    def _get_textFill(self):
        if self._fill is None:
            self._fill = color(r=self.nsColor.redComponent(),
                    g=self.nsColor.greenComponent(),
                    b=self.nsColor.blueComponent(),
                    a=self.nsColor.alphaComponent())
        return self._fill
    textFill = fill = property(_get_textFill)

    # Font stuff

    def _get_displayName(self):
        return self.nsFont.displayName()
    displayName = property(_get_displayName)

    def _get_familyName(self):
        return self.nsFont.familyName()
    familyName = property(_get_familyName)

    def _get_fontName(self):
        return self.nsFont.fontName()
    fontName = font = property(_get_fontName)

    def _get_font(self):
        if self._font is None:
            self._font = findFont(self.fontName)
        return self._font
    font = property(_get_font)

    def _get_isVertical(self):
        return self.nsFont.isVertical()
    isVertical = property(_get_isVertical)

    def _get_isFixedPitch(self):
        return self.nsFont.isFixedPitch()
    isFixedPitch = property(_get_isFixedPitch)

    def _get_boundingRectForFont(self):
        (x, y), (w, h) = self.nsFont.boundingRectForFont()
        return x, y, w, h
    boundingRectForFont = property(_get_boundingRectForFont)

    def _get_renderingMode(self):
        return self.nsFont.renderingMode()
    renderingMode = property(_get_renderingMode)

    # Font metrics, based on self.nsFont. This can be different from
    # self.fontAscencender and self.fontDescender, etc. which are based on the
    # current setting in the FormattedString.

    def _get_ascender(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsFont.ascender()/fontSize, base=fontSize)
    ascender = property(_get_ascender)

    def _get_descender(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsFont.descender()/fontSize, base=fontSize)
    descender = property(_get_descender)

    def _get_capHeight(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsFont.capHeight()/fontSize, base=fontSize)
    capHeight = property(_get_capHeight)

    def _get_xHeight(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsFont.xHeight()/fontSize, base=fontSize)
    xHeight = property(_get_xHeight)

    def _get_italicAngle(self):
        return self.nsFont.italicAngle()
    italicAngle = property(_get_italicAngle)

    def _get_fontSize(self):
        return pt(self.nsFont.pointSize())
    fontSize = property(_get_fontSize)

    #def _get_leading(self):
    #    return self.nsFont.leading()
    #leading = property(_get_leading)

    def _get_fontMatrix(self):
        return self.nsFont.matrix()
    fontMatrix = property(_get_fontMatrix)

    def _get_textTransform(self):
        return self.nsFont.textTransform()
    textTransform = property(_get_textTransform)

    def _get_underlinePosition(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsFont.underlinePosition()/fontSize, base=fontSize)
    underlinePosition = property(_get_underlinePosition)

    def _get_underlineThickness(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsFont.underlineThickness()/fontSize, base=fontSize)
    underlineThickness = property(_get_underlineThickness)

    #   Paragraph attributes

    def _get_matrix(self):
        return CTRunGetTextMatrix(self._ctRun)
    matrix = property(_get_matrix)

    def _get_alignment(self):
        return self.nsParagraphStyle.alignment()
    alignment = property(_get_alignment)

    def _get_lineSpacing(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsParagraphStyle.lineSpacing()/fontSize, base=fontSize)
    lineSpacing = property(_get_lineSpacing)

    def _get_paragraphSpacing(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsParagraphStyle.paragraphSpacing()/fontSize, base=fontSize)
    paragraphSpacing = property(_get_paragraphSpacing)

    def _get_paragraphSpacingBefore(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsParagraphStyle.paragraphSpacingBefore()/fontSize, base=fontSize)
    paragraphSpacingBefore = property(_get_paragraphSpacingBefore)

    def _get_headIndent(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsParagraphStyle.headIndent()/fontSize, base=fontSize)
    headIndent = property(_get_headIndent)

    def _get_tailIndent(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsParagraphStyle.tailIndent()/fontSize, base=fontSize)
    tailIndent = property(_get_tailIndent)

    def _get_firstLineHeadIndent(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsParagraphStyle.firstLineHeadIndent()/fontSize, base=fontSize)
    firstLineHeadIndent = property(_get_firstLineHeadIndent)

    def _get_lineHeightMultiple(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsParagraphStyle.lineHeightMultiple()/fontSize, base=fontSize)
    lineHeightMultiple = property(_get_lineHeightMultiple)

    def _get_maximumLineHeight(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsParagraphStyle.maximumLineHeight()/fontSize, base=fontSize)
    maximumLineHeight = leading = property(_get_maximumLineHeight)

    def _get_minimumLineHeight(self):
        fontSize = self.nsFont.pointSize()
        return em(self.nsParagraphStyle.minimumLineHeight()/fontSize, base=fontSize)
    minimumLineHeight = property(_get_minimumLineHeight)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# -----------------------------------------------------------------------------
#     Copyright (c) 2016+ Buro Petr van Blokland + Claudia Mens
#     www.pagebot.io
#
#     P A G E B O T
#
#     Licensed under MIT conditions
#
#     Supporting DrawBot, www.drawbot.com
#     Supporting Flat, xxyxyz.org/flat
# -----------------------------------------------------------------------------
#
#     errors.py
#
#

class PageBotOSXError(TypeError):
    pass

class PageBotOSXFileFormatError(Exception):

    def __init__(self, msg):
        super().__init__()
        self.msg = msg

    def __str__(self):
        return '! PageBot OSX file format error: %s' % self.msg



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# -----------------------------------------------------------------------------
#
#     P A G E B O T
#
#     Copyright (c) 2016+ Buro Petr van Blokland + Claudia Mens
#     www.pagebot.io
#     Licensed under MIT conditions
#
#     Supporting DrawBot, www.drawbot.com
#     Supporting Flat, xxyxyz.org/flat
# -----------------------------------------------------------------------------
#
#     drawbotstring.py
#

import re
from AppKit import NSAttributeDictionary, NSRange
from CoreText import (CTFramesetterCreateWithAttributedString,
        CTFramesetterCreateFrame, CTFrameGetLines, CTFrameGetLineOrigins)
from Quartz import CGPathAddRect, CGPathCreateMutable, CGRectMake
from pagebot.constants import LEFT, DEFAULT_FONT_SIZE, DEFAULT_LEADING
from pagebot.contexts.base.babelstring import getLineHeight, BabelString
from pagebot.toolbox.color import color, noColor
from pagebot.toolbox.units import pt, upt, units, em
import drawBot as drawBotBuilder
from pagebotosx.strings.textline import TextLine

def pixelBounds(fs):
    """Answers the pixel-bounds rectangle of the text.

    NOTE that @by can be a negative value, if there is text (e.g. overshoot)
    below the baseline.

    - @bh is the amount of pixels above the baseline.
    - For the total height of the pixel-map, calculate @ph - @py.
    - For the total width of the pixel-map, calculate @pw - @px."""
    if not fs:
        return pt(0, 0, 0, 0)
    p = drawBotBuilder.BezierPath()
    p.text(fs, (0, 0))

    '''
    OS X answers `bw` and `bh` as difference with `bx` and `by`. This is not
    very intuitive; in this situation the the total (width, height) always
    needs to be calculated by the caller. Instead, the width and height
    answered is the complete bounding box, and (x, y) is the position of the
    bounding box, compared to (0, 0) of the string origin.
    '''
    bx, by, bw, bh = p.bounds()
    return pt(bx, by, bw - bx, bh - by)

class DrawBotString(BabelString):
    """DrawBotString is a wrapper around the standard DrawBot FormattedString."""

    BABEL_STRING_TYPE = 'fs'
    formattedAttrKeys = ['font', 'fallbackFont', 'fontSize', 'fill',
            'cmykFill', 'stroke', 'cmykStroke', 'strokeWidth', 'align',
            'lineHeight', 'tracking', 'baselineShift', 'underline',
            'openTypeFeatures', 'fontVariations', 'tabs', 'indent',
            'tailIndent', 'firstLineIndent', 'paragraphTopSpacing',
            'paragraphBottomSpacing', 'language', ]


    def __init__(self, s, context, style=None):
        """Constructor of the DrawBotString, wrapper around DrawBot
        FormattedString. Optionally stores the (latest) style that was used to
        produce the formatted string.

        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> from pagebot.fonttoolbox.objects.font import findFont
        >>> font = findFont('Pagebot-Regular')
        >>> style = dict(font=font, fontSize=pt(80))
        >>> bs = context.newString('Example Text', style=style)
        >>> bs.fontSize, round(upt(bs.xHeight)), bs.xHeight, bs.capHeight, bs.ascender, bs.descender
        (80pt, 42, 0.53em, 0.71em, 0.93em, -0.24em)
        >>> #bs.font # FIXME: returns Roboto instead of PageBot font.
        >>> #'PageBot-Regular.ttf' in bs.font
        #True
        >>> #'/PageBot' in bs.fontPath
        #True
        >>> style = dict(font='Verdana', fontSize=pt(100), leading=em(1.4))
        >>> bs = context.newString('Example Text', style=style)
        >>> from pagebot.contexts.base.babelstring import BabelString
        >>> isinstance(bs, BabelString)
        True
        >>> bs[2:]
        ample Text
        >>> #lines = bs.getTextLines(w=100)
        >>> #lines
        #[<TextLine #0 y:181.00 Runs:1>, <TextLine #1 y:41.00 Runs:1>]
        >>> #len(lines)
        #2
        >>> #line = lines[0]
        >>> #line.xHeight, line.capHeight # Max metrics of all runs in line as Em
        #(0.55em, 0.73em)
        >>> #run = line.textRuns[0]
        >>> #run.xHeight, run.capHeight
        #(0.55em, 0.73em)
        >>> fs = context.newString('blablabla')
        >>> fs[2:]
        ablabla
        >>> fs
        blablabla
        >>> fs[:5]
        blabl
        >>> fs[0]
        b
        >>> fs[5]
        a
        """
        fsAttrs = {}
        # Some checking, in case we get something else here.
        assert style is None or isinstance(style, dict)

        # Optional style to set the context parameters. In case defined, store
        # current status here as property and set the current FormattedString
        # for future additions. Also the answered metrics will not be based on
        # these values.
        if style is None:
            style = {}

        self.style = style
        self.language = style.get('language')

        for k, v in style.items():
            if k  == 'leading':
                fontSize = style.get('fontSize', DEFAULT_FONT_SIZE)
                lineHeight = getLineHeight(v, fontSize)
                fsAttrs['lineHeight'] = lineHeight
            elif k == 'hyphenation':
                # Global, not a FS attribute.
                self.hyphenation = v
            elif k in self.formattedAttrKeys:
                fsAttrs[k] = v
            else:
                print('%s not allowed' % k)

        # Turns plain text string `s` into a FormattedString.
        self.s = context.b.FormattedString(s, **fsAttrs)

        # Sets Variable Font fitting parameters.
        self.fittingFontSize = pt(style.get('fontSize'))
        self.fittingFont = style.get('font')
        self.fittingLocation = style.get('location')
        self.isFitting = True
        super().__init__(context)

    def _get_s(self):
        """Answers the embedded FormattedString using a property to force string
        type checking."""
        return self._s

    def _set_s(self, s):
        """ Checks the type of `s`. Three types are supported here: plain
        strings, DrawBot FormattedString and the class self."""
        assert isinstance(s, (DrawBotString, str)) or s.__class__.__name__ == 'FormattedString'

        if isinstance(s, str):
            # newString() instead to inherit default styles?
            s = self.context.b.FormattedString(s)

        elif isinstance(s, DrawBotString):
            # TODO: needs a deeper copy?
            s = s.s

        self._s = s

    s = property(_get_s, _set_s)

    def columnStart(self, firstColumnIndent):
        bs = self
        style = self.getStyleAtIndex(0)

        # Something going on at start?
        if style.get('firstLineIndent') is not None or firstColumnIndent is not None:

            # Really really small place holder period.
            style['fontSize'] = pt(0.0001)

            # Transparant, so it will never show.
            style['textFill'] = color(1, 1, 1, 1)

            # Then make this one work.
            style['firstLineIndent'] = firstColumnIndent or 0
            bs = self.context.newString('.', style=style) + bs
        return bs

    def _get_font(self):
        """Answers the current state of fontName."""
        return self.style.get('font')

    def _set_font(self, fontName):
        if fontName is not None:
            self.context.font(fontName)
        self.style['font'] = fontName

    font = property(_get_font, _set_font)

    def _get_fontSize(self):
        """Answers the current state of the fontSize.

        >>> from pagebot.toolbox.units import mm
        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> style = dict(font='Verdana', fontSize=pt(85), leading=em(1.4))
        >>> bs = context.newString('Example Text', style=style)
        >>> bs.fontSize
        85pt
        >>> # Auto-convert to points
        >>> bs.fontSize = 96
        >>> bs.fontSize
        96pt
        >>> # Set at unit.
        >>> bs.fontSize = mm(5)
        >>> bs.fontSize
        5mm
        >>> #bs.leading
        >>>
        1.4em
        """
        return units(self.style.get('fontSize'))

    def _set_fontSize(self, fontSize):
        if fontSize is not None:
            self.context.fontSize(upt(fontSize))
        self.style['fontSize'] = fontSize

    fontSize = property(_get_fontSize, _set_fontSize)

    def getStyleAtIndex(self, index):
        """Answers the style dictionary with values at position index of the
        string.

        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> c1 = color(0.2, 0.3, 0.4)
        >>> c2 = color(1, 0, 0.22)
        >>> style = style=dict(font='Verdana', fontSize=pt(17), leading=pt(21), textFill=c1, textStroke=c2, textStrokeWidth=pt(3))
        >>> bs = context.newString('Example Text1', style=style)
        >>> bs.getStyleAtIndex(3)['fontSize']
        17pt
        >>> style = dict(font='Georgia', fontSize=pt(20), leading=pt(25), textFill=c2, textStroke=c1, textStrokeWidth=pt(2))
        >>> bs += context.newString('Another Text', style=style)
        >>> bs.getStyleAtIndex(20)['leading']
        25pt
        >>> bs.getStyleAtIndex(20)['textFill']
        Color(r=1.0, g=0, b=0.22)
        """
        attrString = self.s.getNSObject()
        cIndex = 0
        style = {}

        if attrString:
            for nsObject in attrString.attributesAtIndex_effectiveRange_(index, None):
                if isinstance(nsObject, NSAttributeDictionary):
                    nsColor = nsObject.get('NSColor')
                    if nsColor is not None:
                        style['textFill'] = color(nsColor.redComponent(), nsColor.greenComponent(), nsColor.blueComponent())

                    nsColor = nsObject.get('NSStrokeColor')

                    if nsColor is not None:
                        style['textStroke'] = color(nsColor.redComponent(), nsColor.greenComponent(), nsColor.blueComponent())

                    strokeWidth = nsObject.get('NSStrokeWidth')

                    if strokeWidth is not None:
                        style['strokeWidth'] = pt(strokeWidth)

                    nsFont = nsObject.get('NSFont')

                    if nsFont is not None:
                        style['font'] = nsFont.fontName()
                        style['fontSize'] = pt(nsFont.pointSize())

                    pgStyle = nsObject.get('NSParagraphStyle')
                    style['tabs'] = tabs = {}
                    #for tab in pgStyle.tabStops:
                    #    tabe[tab] = 'a'
                    style['leading'] = pt(pgStyle.minimumLineHeight())
                    style['firstLineIndent'] = pt(pgStyle.firstLineHeadIndent())
                    style['indent'] = pt(pgStyle.headIndent())
                    style['tailIndent'] = pt(pgStyle.tailIndent())
                    # MORE FROM: Alignment 4, LineSpacing 0, ParagraphSpacing 0, ParagraphSpacingBefore 0,
                    #  0, LineHeight 21/21, LineHeightMultiple 0, LineBreakMode 0, Tabs (), DefaultTabInterval 0,
                    # Blocks (), Lists (), BaseWritingDirection -1, HyphenationFactor 0, TighteningForTruncation NO, HeaderLevel 0
                elif isinstance(nsObject, NSRange):
                    if cIndex >= index: # Run through until matching index, so the style cumulates.
                        break
                    cIndex += nsObject.length
        return style

    def asText(self):
        """Answers the text string.

        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> bs = context.newString('Example Text')
        >>> bs.asText()
        'Example Text'
        """
        return str(self.s) #  Convert to text

    def textSize(self, w=None, h=None):
        """Answers the (w, h) size for a given width, with the current text,
        measured from bottom em-size to top em-size (including ascender+ and
        descender+) and the string width (including margins).

        >>> from pagebot.toolbox.units import mm, uRound
        >>> from pagebot import getContext
        >>> from pagebot.fonttoolbox.objects.font import findFont
        >>> context = getContext('DrawBot')
        >>> font = findFont('Bungee-Regular')
        >>> style = dict(font=font, fontSize=pt(12))
        >>> bs = context.newString('Example Text ' * 20, style=style)
        >>> #len(bs.getTextLines(w=100))
        #16
        >>> uRound(bs.textSize(w=300))
        [290pt, 130pt]
        """
        b = self.context.b

        if w is not None:
            wpt = upt(w)
            return b.textSize(self.s, width=wpt)
        if h is not None:
            hpt = upt(h)
            return b.textSize(self.s, height=hpt)

        return b.textSize(self.s)

    def bounds(self):
        """Answers the pixel-bounds rectangle of the text, if formatted by the
        option (w, h).

        NOTE that @by can be a negative value, if there is text (e.g.
        overshoot) below the baseline.
        @bh is the amount of pixels above the baseline.
        For the total height of the pixel-map, calculate @ph - @py.
        For the total width of the pixel-map, calculate @pw - @px."""

        # Sets the hyphenation flag and language from self; in DrawBot this
        # is set by a global function, not as FormattedString attribute.
        self.context.language(self.language)
        self.context.hyphenation(self.hyphenation)
        return pixelBounds(self.s)

    def fontContainsCharacters(self, characters):
        """Return a Boolean if the current font contains the provided
        characters. Characters is a string containing one or more
        characters."""
        return self.s.fontContainsCharacters(characters)

    def _get_fontFilePath(self):
        """Return the path to the file of the current font."""
        return self.s.fontFilePath()

    fontPath = property(_get_fontFilePath)

    def listFontGlyphNames(self):
        """Return a list of glyph names supported by the current font."""
        return self.s.listFontGlyphNames()

    def _get_ascender(self):
        """Returns the current font ascender as relative Em, based on the
        current font and fontSize."""
        fontSize = upt(self.fontSize)
        return em(self.s.fontAscender() / fontSize, base=fontSize)

    # Compatibility with DrawBot API.
    fontAscender = ascender = property(_get_ascender)

    def _get_descender(self):
        """Returns the current font descender as Em, based on the current font
        and fontSize."""
        fontSize = upt(self.fontSize)
        return em(self.s.fontDescender()/fontSize, base=fontSize)

    # Compatibility with DrawBot API.
    fontDescender = descender = property(_get_descender)

    def _get_xHeight(self):
        """Returns the current font x-height as Em, based on the current font
        and fontSize."""
        fontSize = upt(self.fontSize)
        return em(self.s.fontXHeight()/fontSize, base=fontSize)

    # Compatibility with DrawBot API.
    fontXHeight = xHeight = property(_get_xHeight)

    def _get_capHeight(self):
        """Returns the current font cap height as Em, based on the current font
        and fontSize."""
        fontSize = upt(self.fontSize)
        return em(self.s.fontCapHeight()/fontSize, base=fontSize)

    # Compatibility with DrawBot API.
    fontCapHeight = capHeight = property(_get_capHeight)

    def _get_leading(self):
        """Answers the current leading value."""
        #fontSize = upt(self.fontSize)
        #return em(self.s.fontLeading() / fontSize, base=fontSize)

        return self.style.get('leading', DEFAULT_LEADING)

    # Compatibility with DrawBot API.
    fontLeading = leading = property(_get_leading)

    def _get_lineHeight(self):
        """Returns the current line height, based on the current font and fontSize.
        If a lineHeight is set, this value will be returned."""
        #fontSize = upt(self.fontSize)
        return self.s.fontLineHeight()

    # Compatibility with DrawBot API.
    fontLineHeight = lineHeight = property(_get_lineHeight)

    def appendGlyph(self, *glyphNames):
        """Append a glyph by his glyph name using the current font. Multiple
        glyph names are possible."""
        self.s.appendGlyph(glyphNames)

    MARKER_PATTERN = '==%s@%s=='
    FIND_FS_MARKERS = re.compile('\=\=([a-zA-Z0-9_\:\.]*)\@([^=]*)\=\=')

    def appendMarker(self, markerId, arg):
        """Appends a formatted string with markerId that can be used as
        non-display marker. This way the Composer can find the position of
        markers in text boxes, after FS-slicing has been done. Note there is
        always a very small "white-space" added to the string, so there is a
        potential difference in width that matters. For that reason markers
        should not be changed after slicing (which would theoretically alter
        the flow of the FormattedString in an box) and the markerId and
        amount/length of args should be kept as small as possible.

        NOTE that there is a potential problem of slicing through the argument
        string at the end of a textBox. That is another reason to keep the
        length of the arguments short. And not to use any spaces, etc. inside
        the markerId. Possible slicing through line-endings is not a problem,
        as the raw string ignores them."""
        marker = self.MARKER_PATTERN % (markerId, arg or '')
        fs = self.context.b.FormattedString(marker, fill=noColor,
                stroke=noColor, fontSize=0.0000000000001)
        self.append(fs)

    def findMarkers(self, reCompiled=None):
        """Answers a dictionary of markers with their arguments in self.s."""
        if reCompiled is None:
            reCompiled= self.FIND_FS_MARKERS
        return reCompiled.findall(u'%s' % self.s)

    def textOverflow(self, w, h, align=LEFT):
        """Answers the overflowing of from the box (0, 0, w, h) as a new
        DrawBotString in the current context."""
        wpt, hpt = upt(w, h)
        box = (0, 0, wpt, hpt)

        # Sets the hyphenation flag from style. In DrawBot this is set by a
        # global function, not as FormattedString attribute.

        # TODO: Attributes don't seem to stay the same in the string or overfill copy.
        language = self.language or 'en'
        hyphenation = self.hyphenation or True
        self.context.b.language(language)
        self.context.b.hyphenation(hyphenation)
        overflow = self.__class__(self.context.b.textOverflow(self.s, box, align), self.context)

        # Pass on these parameters to the new constructed DrawBotString.
        overflow.language = language
        overflow.hyphenation = hyphenation
        return overflow

    def getBaselines(self, w, h=None):
        """Answers the dictionary of vertical baseline positions for the self.s
        FormattedString and for the given width and height. The value is the
        TextLine instance at that position.

        FIXME

        >>> from pagebot.toolbox.units import mm, uRound
        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> style = dict(font='Verdana', fontSize=pt(12))
        >>> bs = context.newString('Example Text ' * 10, style=style)
        >>> baselines = bs.getBaselines(w=200)
        >>> #baselines # FIXME: y-values too large
        """
        baselines = {}

        for textLine in self.getTextLines(w, h):
            baselines[textLine.y.pt] = textLine

        return baselines

    def getTextLines(self, w, h=None, align=LEFT):
        """Answers the dictionary of TextLine instances. Key is y position of
        the line.

        >>> from pagebot.toolbox.units import mm, uRound
        >>> from pagebot import getContext
        >>> from pagebot.fonttoolbox.objects.font import findFont
        >>> context = getContext('DrawBot')
        >>> font = findFont('Bungee-Regular')
        >>> style = dict(font=font, fontSize=pt(12))
        >>> bs = context.newString('Example Text', style=style)
        >>> lines = bs.getTextLines(w=200, h=200)
        >>> len(lines)
        1
        >>> #lines # FIXME: baseline shift on Travis OSX.
        #[<TextLine #0 y:185.20 Runs:1>]
        >>> line = lines[0]
        >>> line.maximumLineHeight
        1.4em
        >>> #line.y # FIXME: baseline shift on Travis OSX.
        #185.2pt
        >>> #lines = bs.getTextLines(w=200, h=200)
        >>> attrString = bs.s.getNSObject()
        >>> len(attrString)
        12
        >>> setter = CTFramesetterCreateWithAttributedString(attrString)
        >>> path = CGPathCreateMutable()
        >>> CGPathAddRect(path, None, CGRectMake(0, 0, 200, 600))
        >>> ctBox = CTFramesetterCreateFrame(setter, (0, 0), path, None)
        >>> ctLines = CTFrameGetLines(ctBox)
        >>> from CoreText import CTLineGetGlyphRuns, CTRunGetAttributes
        >>> runs = CTLineGetGlyphRuns(ctLines[0])
        >>> attrs = CTRunGetAttributes(runs[0])
        >>> origins = CTFrameGetLineOrigins(ctBox, (0, len(ctLines)), None)
        >>> lineHeight = 16.8
        >>> oy = origins[0].y
        >>> #oy # FIXME: baseline shift on Travis OSX.
        #585.2
        >>> 600 - lineHeight
        583.2
        """
        assert w

        if h is None:
            h = 3 * w

        wpt, hpt = upt(w, h)
        textLines = []
        attrString = self.s.getNSObject()
        setter = CTFramesetterCreateWithAttributedString(attrString)
        path = CGPathCreateMutable()
        CGPathAddRect(path, None, CGRectMake(0, 0, wpt, hpt))
        ctBox = CTFramesetterCreateFrame(setter, (0, 0), path, None)
        ctLines = CTFrameGetLines(ctBox)
        origins = CTFrameGetLineOrigins(ctBox, (0, len(ctLines)), None)

        for lIndex, ctLine in enumerate(ctLines):
            origin = origins[lIndex]
            origin_y = origin.y
            textLine = TextLine(ctLine, pt(origin.x), pt(origin_y), lIndex)
            textLines.append(textLine)

        return textLines

    @classmethod
    def newString(cls, s, context, e=None, style=None, w=None, h=None, **kwargs):
        """Answers a DrawBotString instance from valid attributes in *style*.
        Set all values after testing their existence, so they can inherit from
        previous style formats in the string.

        If target width *w* or height *h* is defined, then *fontSize* is scaled
        to make the string fit *w* or *h*.
        TODO: restore pixelFit implementation.

        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> from pagebot.fonttoolbox.objects.font import findFont
        >>> font = findFont('Roboto-Regular')
        >>> font = findFont('Bungee-Regular')
        >>> font = findFont('Roboto-Black')
        >>> bs = context.newString('ABC', style=dict(font=font.path, fontSize=pt(22)))
        >>> bs.style['font'].endswith('Roboto-Black.ttf')
        True
        >>> bs
        ABC
        >>> bs = context.newString('ABC', style=dict(font=font.path, w=pt(100)))
        >>> int(round(bs.fontSize))
        12
        >>> # Use the font instance instead of path.
        >>> bs = context.newString('ABC ', style=dict(font=font, w=pt(100)))
        >>> int(round(bs.fontSize))
        12
        >>> bs1 = context.newString('DEF')
        >>> bs + bs1
        ABC DEF
        """
        assert isinstance(s, str)

        if style is None:
            style = {}
        else:
            assert isinstance(style, dict)

        attrs = cls.getStringAttributes(s, e=e, style=style, w=w, h=h)
        s = cls.addCaseToString(s, e, style)
        # Creates a new DrawBotString.
        return cls(s, context, style=attrs)

if __name__ == '__main__':
    import doctest
    import sys
    sys.exit(doctest.testmod()[0])



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# -----------------------------------------------------------------------------
#
#     P A G E B O T
#
#     Copyright (c) 2016+ Buro Petr van Blokland + Claudia Mens
#     www.pagebot.io
#     Licensed under MIT conditions
#
#     Supporting DrawBot, www.drawbot.com
#     Supporting Flat, xxyxyz.org/flat
# -----------------------------------------------------------------------------
#
#     drawbotcontext.py
#

import os
from CoreText import (CTFontDescriptorCreateWithNameAndSize,
        CTFontDescriptorCopyAttribute, kCTFontURLAttribute)
from AppKit import NSFont
import drawBot
from pagebot.constants import *
from pagebot.contexts.base.babelstring import BabelString
from pagebot.contexts.base.basecontext import BaseContext
from pagebot.toolbox.color import color, noColor
from pagebot.toolbox.units import pt, upt, point2D
from pagebot.toolbox.transformer import path2Name, path2Dir
from drawBot import Variable
from pagebotosx.contexts.drawbot.drawbotstring import DrawBotString as stringClass
#from pagebotosx.bezierpaths.osxbezierpath import OSXBezierPath

# Identifier to make builder hook name. Views will try to call e.build_html()
drawBotBuilder = drawBot
drawBotBuilder.PB_ID = 'drawBot'

class DrawBotContext(BaseContext):
    """DrawBotContext adapts DrawBot functionality to PageBot."""

    STRING_CLASS = stringClass
    EXPORT_TYPES = (FILETYPE_PDF, FILETYPE_SVG, FILETYPE_PNG, FILETYPE_JPG,
            FILETYPE_GIF, FILETYPE_MOV)

    '''
    /_scaled will be ignored with default .gitignore settings.  If the
    docs/images/_scaled folder need to be committed to a Git repository, remove
    _scaled from .gitignore.
    '''
    SCALED_PATH = '_scaled' # /scaled with upload on Git. /_scaled will be ignored.

    def __init__(self):
        """Constructor of DrawBotContext if drawBot import exists.

        >>> drawBotBuilder is not None
        True
        >>> drawBotBuilder is not None and drawBotBuilder.PB_ID == 'drawBot'
        True
        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> context.name
        'DrawBotContext'
        """
        super().__init__()
        # The context builder "cls.b" is drawBot which executes actual drawing
        # calls, similar to function calls in DrawBot scripts.
        self.b = drawBotBuilder #  Builder for this canvas.
        self.name = self.__class__.__name__
        # Holds the extension as soon as the export file path is defined.
        self.fileType = DEFAULT_FILETYPE

    # Drawing.

    def newDrawing(self, w=None, h=None, doc=None):
        """Clear output canvas, start new export file. DrawBot function.

        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> context.newDrawing()
        """
        self.b.newDrawing()

    def endDrawing(self, doc=None):
        self.b.endDrawing()

    def saveDrawing(self, path, multiPage=None):
        """Select non-standard DrawBot export builders here. Save the current
        image as path, rendering depending on the extension of the path file.
        In case the path starts with "_export", create its directories.

        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> context.saveImage('_export/MyFile.pdf')

        """
        self.checkExportPath(path)

        if path.lower().endswith('.mov'):
            self.b.saveImage(path)
        else:
            self.b.saveImage(path, multipage=multiPage)

    saveImage = saveDrawing

    def getDrawing(self):
        """Returns a PDF document of the current state."""
        return self.b.pdfImage()

    def setStyles(self, styles):
        pass

    #   V A R I A B L E

    def Variable(self, variables, workSpace):
        """Offers interactive global value manipulation in DrawBot. Should be
        ignored in other contexts.

        Variable is a DrawBot context global, used to make simple UI with
        controls on input parameters."""
        Variable(variables, workSpace)

    #   D R A W I N G

    def bluntCornerRect(self, x, y, w, h, offset=5):
        """Draw a rectangle in the canvas. This method is using the core
        BezierPath as path to draw on. For a more rich environment use
        PageBotPath(context) instead.

        TODO: move to elements.

        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> context.bluntCornerRect(pt(0), pt(0), pt(100), pt(100))
        >>> context.bluntCornerRect(0, 0, 100, 100)
        """

        xPt, yPt, wPt, hPt, offsetPt = upt(x, y, w, h, offset)
        path = self.newPath() #
        path.moveTo((xPt+offsetPt, yPt))
        path.lineTo((xPt+wPt-offsetPt, yPt))
        path.lineTo((xPt+wPt, yPt+offsetPt))
        path.lineTo((xPt+wPt, yPt+hPt-offsetPt))
        path.lineTo((xPt+w-offsetPt, y+hPt))
        path.lineTo((xPt+offsetPt, y+hPt))
        path.lineTo((xPt, yPt+h-offsetPt))
        path.lineTo((xPt, yPt+offsetPt))
        self.closePath()
        self.drawPath(path)

    def roundedRect(self, x, y, w, h, offset=25):
        """Draw a rectangle in the canvas. This method is using the core BezierPath
        as path to draw on. For a more rich environment use PageBotPath(context)
        instead.

        TODO: move to elements.

        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> context.roundedRect(pt(0), pt(0), pt(100), pt(100))
        >>> context.roundedRect(0, 0, 100, 100)
        """
        xPt, yPt, wPt, hPt, offsetPt = upt(x, y, w, h, offset)
        path = self.newPath()
        path.moveTo((xPt+offsetPt, yPt))
        path.lineTo((xPt+wPt-offsetPt, yPt))
        path.curveTo((xPt+wPt, yPt), (xPt+wPt, yPt), (xPt+wPt, yPt+offsetPt))
        path.lineTo((xPt+wPt, yPt+hPt-offsetPt))
        path.curveTo((xPt+wPt, yPt+hPt), (xPt+wPt, yPt+hPt), (xPt+wPt-offsetPt, yPt+hPt))
        path.lineTo((xPt+offsetPt, yPt+hPt))
        path.curveTo((xPt, yPt+hPt), (xPt, yPt+hPt), (xPt, yPt+hPt-offsetPt))
        path.lineTo((xPt, yPt+offsetPt))
        path.curveTo((xPt, yPt), (xPt, yPt), (xPt+offsetPt, yPt))
        self.closePath()
        self.drawPath(path)

    #   T E X T
    #

    def text(self, s, p):
        """Draws the s text string at position p.

        `S` Can be a str, BabelString, or DrawBot FormattedString.

        NOTE: signature differs from DrawBot.
        """
        if isinstance(s, BabelString):
            # BabelString with a FormattedString inside.
            s = s.s
            position = point2D(upt(p))
        else:
            if not isinstance(s, str):
                # Otherwise converts to string if it is not already.
                s = str(s)
            # Regular string, use global font and size.
            position = point2D(upt(p))
            self.b.fontSize(self._fontSize)
            self.font(self._font)

        # Render point units to value tuple
        self.b.text(s, position)

    def textBox(self, s, r=None, clipPath=None, align=None):
        """Draws the s text string in rectangle r.

        `S` Can be a str, BabelString, or DrawBot FormattedString.

        NOTE: signature differs from DrawBot.

        >>> from pagebot.toolbox.units import pt
        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> context.newPage(420, 420)
        >>> txt = '''The 25-storey Jumeirah Beach Hotel, with its distinctive\
design in the shape of a wave, has become one of the most successful\
hotels in the world. Located on Jumeirah Beach, this well-known hotel\
offers a wonderful holiday experience and a variety of pleasurable\
activities. The many restaurants, bars and cafés, daily live\
entertainment and sports facilities will keep you entertained, whilst\
children will have a great time at the Sinbad’s Kids’ Club or Wild Wadi\
WaterparkTM which is freely accessible through a private gate.'''
        >>> bs = context.newString(txt)
        >>> context.fontSize(14)
        >>> tb = context.textBox(txt, r=(100, 450, 200, 300))
        >>> tb = context.textBox(bs, r=(100, 450, 200, 300))
        """
        tb = None

        if hasattr(s, 's'):
            # Assumes it's a BabelString with a FormattedString inside.
            s = s.s
        else:
            if not isinstance(s, str):
                # Otherwise converts to string if it is not already.
                s = str(s)

        if clipPath is not None:
            box = clipPath.bp
            tb = self.b.textBox(s, clipPath.bp)
        elif isinstance(r, (tuple, list)):
            # Renders rectangle units to value tuple.
            xpt, ypt, wpt, hpt = upt(r)
            box = (xpt, ypt, wpt, hpt)
            tb = self.b.textBox(s, box, align=None)
        else:
            msg = '%s.textBox has no box or clipPath defined' % self.__class__.__name__
            raise ValueError(msg)

        return tb

    def textOverflow(self, s, box, align=None):
        """Answers the overflow text if flowing it in the box. In case a plain
        string is given then the current font / fontSize / ... settings of the
        builder are used.

        `S` Can be a str, BabelString, or DrawBot FormattedString.

        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> context.newDrawing()
        >>> context.newPage(420, 420)
        >>> context.font('Verdana')
        >>> context.fontSize(12)
        >>> box = 0, 0, 300, 20
        >>> s = 'AAA ' * 200
        >>> len(s)
        800
        >>> # Plain string overflow.
        >>> of = context.textOverflow(s, box)
        >>> len(of)
        760
        >>> # Styled DrawBotString overflow.
        >>> style = dict(font='Verdana', fontSize=14)
        >>> bs = context.newString('AAA ' * 200, style=style)
        >>> of = context.textOverflow(bs, box)
        >>> len(of)
        764
        """
        if isinstance(s, str):
            return self.b.textOverflow(s, box, align=align) # Plain string

        # Assume here it's a DrawBotString with a FormattedString inside
        overflow = self.b.textOverflow(s.s, box, align=align)
        #bs = self.newString('')
        #bs.s = overflow
        return overflow

    def textBoxBaselines(self, txt, box, align=None):
        return self.b.textBoxBaselines(txt, box, align=align)

    def textSize(self, bs, w=None, h=None, align=None):
        """Answers the width and height of the formatted string with an
        optional given w or h."""
        return self.b.textSize(bs.s, width=w, height=h, align=align)

    def hyphenation(self, onOff):
        self.b.hyphenation(onOff)

    def language(self, language):
        self.b.language(language)

    #   P A T H
    #
    #   Function that work on the current running path stored in self._bezierpath
    #

    def newPath(self):
        #self._bezierpath = OSXBezierPath(self.b)
        # Using DrawBot's BezierPath class for now.
        self._bezierpath = self.b.BezierPath()
        return self.bezierpath

    def drawGlyphPath(self, glyph):
        """Converts the cubic commands to a drawable path."""
        path = self.getGlyphPath(glyph)
        self.drawPath(path)

    def getGlyphPath(self, glyph, p=None, path=None):
        """Answers the DrawBot path. Allow optional position offset and path,
        in case we do recursive component drawing.

        >>> from pagebot.fonttoolbox.objects.font import findFont
        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> f = findFont('Roboto-Regular')
        >>> print(f)
        <Font Roboto-Regular>
        >>> g = f['H']
        >>> path = context.getGlyphPath(g)
        """
        if path is None:
            path = self.newPath()

        if p is None:
            px = py = 0
        else:
            px = p[0]
            py = p[1]

        for command, t in glyph.cubic:
            if command == 'moveTo':
                path.moveTo((px+t[0], py+t[1]))
            elif command == 'lineTo':
                path.lineTo((px+t[0], py+t[1]))
            elif command == 'curveTo':
                path.curveTo((px+t[0][0], py+t[0][1]),
                        (px+t[1][0], py+t[1][1]), (px+t[2][0], py+t[2][1]))
            elif command == 'closePath':
                path.closePath()
            elif command == 'component':
                (x, y), componentGlyph = t
                self.getGlyphPath(componentGlyph, (px+x, py+y), path)

        return path

    def bezierPathByFlatteningPath(self, path=None):
        """Use the NSBezierPath flatten function. Answers None if the flattened
        path could not be made.
        """
        if path is None:
            path = self.bezierpath

        if hasattr(path, 'path'):
            # In case it is a PageBotPath.
            path = path.path

        return path._path.bezierPathByFlatteningPath()

    def getFlattenedPath(self, path=None):
        return self.bezierPathByFlatteningPath(path=path)

    def getFlattenedContours(self, path=None):
        """Answers the flattened Bézier path as  a contour list [contour,
        contour, ...] where contours are lists of point2D() points.

        """
        contour = []
        flattenedContours = [contour]

        # Use / create self._bezierpath if path is None.
        flatPath = self.bezierPathByFlatteningPath(path)

        if flatPath is not None:
            for index in range(flatPath.elementCount()):
                # NSBezierPath size + index call.
                p = flatPath.elementAtIndex_associatedPoints_(index)[1]

                if p:
                    # Make point2D() tuples, no need to add point type, all
                    # onCurve.
                    contour.append((p[0].x, p[0].y))
                else:
                    contour = []
                    flattenedContours.append(contour)

        return flattenedContours

    # Path drawing behavior.

    def strokeWidth(self, w):
        """Set the current stroke width.

        >>> from pagebot.toolbox.units import pt, mm
        >>> from pagebot import getContext
        >>> context = getContext()
        >>> context.newDrawing()
        >>> context.newPage(420, 420)
        >>> context.setStrokeWidth(pt(0.5))
        >>> context.setStrokeWidth(mm(0.5))
        """
        wpt = upt(w)
        self.b.strokeWidth(wpt)

    setStrokeWidth = strokeWidth

    def miterLimit(self, value):
        self.b.miterLimit(value)

    def lineJoin(self, value):
        self.b.lineJoin(value)

    def lineCap(self, value):
        """Possible values are butt, square and round."""
        assert value in ('butt', 'square', 'round')
        self.b.lineCap(value)

    def lineDash(self, value):
        """LineDash is None or a list of dash lengths."""
        if value is None:
            self.b.lineDash(None)
        else:
            self.b.lineDash(*value)

    #   F O N T S

    def fontPath2FontName(self, fontPath):
        """Answers the font name of the font related to fontPath. This is done
        by installing it (again). Answers None if the font cannot be installed
        or if the path does not exists.

        >>> import os
        >>> from pagebot.fonttoolbox.fontpaths import TEST_FONTS_PATH
        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> # Does not exist.
        >>> context.fontPath2FontName('Aaa.ttf') is None
        True
        >>> path = TEST_FONTS_PATH + '/fontbureau/Amstelvar-Roman-VF.ttf'
        >>> os.path.exists(path)
        True
        >>> context.fontPath2FontName(path)
        'Amstelvar-Roman'
        """
        if os.path.exists(fontPath):
            return self.b.font(fontPath)
        return None

    def fontName2FontPath(self, fontName):
        """Answers the unchanged path, if it exists as file. Answers the path
        that is source of the given font name. Answers None if the font cannot
        be found."""
        # If the font cannot be found by name, then test if the file exists as
        # path and answer it.
        if os.path.exists(fontName):
            return fontName

        # Otherwise try OSX for the conversion.
        nsFont = NSFont.fontWithName_size_(fontName, 25)

        if nsFont is not None:
            fontRef = CTFontDescriptorCreateWithNameAndSize(nsFont.fontName(), nsFont.pointSize())
            url = CTFontDescriptorCopyAttribute(fontRef, kCTFontURLAttribute)
            return url.path()
        return None

    #   G L Y P H

    def drawGlyph(self, glyph, x, y, fill=noColor, stroke=noColor,
            strokeWidth=0, fontSize=None, xAlign=CENTER):
        """Draw the font[glyphName] at the defined position with the defined
        fontSize."""
        font = glyph.font

        if fontSize is None:
            fontSize = font.info.unitsPerEm
        s = fontSize/font.info.unitsPerEm

        if xAlign == CENTER:
            x -= (glyph.width or 0)/2*s
        elif xAlign == RIGHT:
            x -= glyph.width*s

        self.save()
        self.fill(fill)
        self.stroke(stroke, w=strokeWidth)
        self.translate(x, y)
        self.scale(s)
        self.drawGlyphPath(glyph)
        self.restore()

    #   I M A G E

    def image(self, path, p=None, alpha=1, pageNumber=None, w=None, h=None,
            scaleType=None):
        """Draws the image. If w or h is defined, scale the image to fit."""
        if p is None:
            p = ORIGIN

        iw, ih = self.imageSize(path)

        if not w and not h:
            w = iw
            h = ih
            sx = sy = 1
        elif scaleType == SCALE_TYPE_FITWH:
            sx = upt(w/iw)
            sy = upt(h/ih)
        elif scaleType == SCALE_TYPE_FITW:
            sx = sy = upt(w/iw)
        elif scaleType == SCALE_TYPE_FITH:
            sx = sy = upt(h/ih)
        else:
            # scaleType in (None, SCALE_TYPE_PROPORTIONAL):
            sx = sy = min(pt(w/iw), upt(h/ih))

        # Else both w and h are defined, scale disproportionally.
        xpt, ypt, = point2D(p)
        self.save()
        self.scale(sx, sy)
        self.translate(xpt/sx, ypt/sy)
        self.b.image(path, (0, 0), alpha=alpha, pageNumber=pageNumber)
        self.restore()

    def ImageObject(self, path=None):
        """Answers an ImageObject that knows about filters. For names
        and parameters of filters see:

        * http://www.drawbot.com/content/image/imageObject.html

        >>> from pagebot import getResourcesPath
        >>> from pagebot import getContext
        >>> context = getContext()
        >>> path = getResourcesPath() + '/images/peppertom_lowres_398x530.png'
        >>> imo = context.getImageObject(path)

        """
        return self.b.ImageObject(path=path)

    def path2ScaledImagePath(self, path, w, h, index=None, exportExtension=None):
        """Answers the path to the scaled image.

        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> context.path2ScaledImagePath('/xxx/yyy/zzz/This.Is.An.Image.jpg', 110, 120)
        ('/xxx/yyy/zzz/_scaled/', 'This.Is.An.Image.110x120.0.jpg')

        >>> path, fileName = context.path2ScaledImagePath('/xxx/yyy/zzz/This.Is.An.Image.jpg', 110, 120)
        >>> path in ('/xxx/yyy/zzz/scaled/', '/xxx/yyy/zzz/_scaled/')
        True
        >>> fileName
        'This.Is.An.Image.110x120.0.jpg'
        """
        # /_scaled will be ignored with default .gitignore settings.
        # If docs/images/_scaled need to be committed into Git repo,
        # then remove _scaled from .gitignore.
        cachePath = '%s/%s/' % (path2Dir(path), self.SCALED_PATH)
        fileNameParts = path2Name(path).split('.')

        # If undefined, take the original extension for exporting the cache.
        if not exportExtension:
            exportExtension = fileNameParts[-1].lower()
        cachedFileName = '%s.%dx%d.%d.%s' % ('.'.join(fileNameParts[:-1]), w, h, index or 0, exportExtension)
        return cachePath, cachedFileName

    def scaleImage(self, path, w, h, index=None, showImageLoresMarker=False,
            exportExtension=None, force=False):
        """Scales the image at the path into a new cached image file. Ignore if
        the cache file is already there.

        First create the new file name, depending on the resolution of the
        scaled image.  Note that in DrawBot this scaling and saving should be
        done before any real document/page drawing started, since this proces
        is using DrawBot canvas pages to execute.

        In case the source contains indexed pages, use index to select the
        page. If omitted, the default index is 0 (in DrawBot this also works on
        non-PDF files).

        >>> from pagebot import getResourcesPath
        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> path = getResourcesPath() + '/images/peppertom_lowres_398x530.png'
        >>> scaledImagePath = context.scaleImage(path, 300, 400)
        >>> os.path.exists(scaledImagePath)
        True
        >>> scaledImagePath = context.scaleImage(path, 3, 4) # Reall small
        >>> os.path.exists(scaledImagePath)
        True
        """
        # If default _scaled directory does not exist, then create it.
        cachePath, fileName = self.path2ScaledImagePath(path, w, h, index, exportExtension)
        if not os.path.exists(cachePath):
            os.makedirs(cachePath)
        cachedFilePath = cachePath + fileName

        if force or not os.path.exists(cachedFilePath):
            # Clean the drawing stack.
            self.newDrawing()
            self.newPage(w=w, h=h)
            self.image(path, (0, 0), w=w, h=h, pageNumber=index or 0)
            if showImageLoresMarker:
                bs = self.newString('LO-RES',
                        style=dict(font=DEFAULT_FALLBACK_FONT_PATH,
                            fontSize=pt(64), fill=color(0, 1, 1),
                            textFill=color(1, 0, 0)))
                tw, th = bs.size
                self.text(bs, (w/2-tw/2, h/2-th/4))
            self.saveImage(cachedFilePath)

            # Clean the drawing stack again.
            self.newDrawing()
        return cachedFilePath

    def imagePixelColor(self, path, p=None):
        if p is None:
            p = ORIGIN
        ppt = point2D(upt(p))
        return self.b.imagePixelColor(path, ppt)

    def numberOfImages(self, path):
        """Answers the number of images in the file referenced by path."""
        return self.b.numberOfPages(path)

    def translate(self, tx, ty):
        self.b.translate(tx, ty)

    def scale(self, sx=1, sy=None, center=(0, 0)):
        if sy is None:
            sy = sx
        self.b.scale(sx, sy, center=center)

    # System fonts listing, installation, font properties.

    def installedFonts(self, patterns=None):
        """Answers the list of all fonts (name or path) that are installed on
        the OS.

        >>> from pagebot import getContext
        >>> context = getContext()
        >>> installed = context.installedFonts()
        >>> len(installed) > 0
        True
        """
        # In case it is a string, convert to a list.
        if isinstance(patterns, str):
            patterns = [patterns]

        fontNames = []

        for fontName in self.b.installedFonts():
            if not patterns:
                # If no pattern then answer all.
                fontNames.append(fontName)
            else:
                for pattern in patterns:
                    if pattern in fontName:
                        fontNames.append(fontName)
                        break
        return fontNames

    def installFont(self, fontOrName):
        """Install the font in the context. fontOrName can be a Font instance
        (in which case the path is used) or a full font path.

        >>> from pagebot.fonttoolbox.objects.font import findFont
        >>> from pagebot import getContext
        >>> context = getContext()
        >>> installed = context.installedFonts()
        >>> len(installed) > 0
        True
        >>> font = findFont('Roboto-Regular')
        >>> context.installFont(font)
        'Roboto-Regular'
        """
        if hasattr(fontOrName, 'path'):
            fontOrName.info.installedName = self.b.installFont(fontOrName.path)
            return fontOrName.info.installedName
        return self.b.installFont(fontOrName)

    def uninstallFont(self, fontOrName):
        if hasattr(fontOrName, 'path'):
            fontOrName = fontOrName.path
        return self.b.uninstallFont(fontOrName)

    def fontContainsCharacters(self, characters):
        return self.b.fontContainsCharacters(characters)

    def fontContainsGlyph(self, glyphName):
        return self.b.fontContainsGlyph(glyphName)

    def fontFilePath(self):
        return self.b.fontFilePath()

    def listFontGlyphNames(self):
        return self.b.listFontGlyphNames()

    def fontAscender(self):
        return self.b.fontAscender()

    def fontDescender(self):
        return self.b.fontDescender()

    def fontXHeight(self):
        return self.b.fontXHeight()

    def fontCapHeight(self):
        return self.b.fontCapHeight()

    def fontLeading(self):
        return self.b.fontLeading()

    def fontLineHeight(self):
        return self.b.fontLineHeight()

    # Features.

    def openTypeFeatures(self, features):
        """Enables OpenType features and returns the current openType features
        settings. If no arguments are given `openTypeFeatures()` will just
        return the current openType features settings.

        In DrawBot:

            size(1000, 200)
            # create an empty formatted string object
            t = FormattedString()
            # set a font
            t.font("ACaslonPro-Regular")
            # set a font size
            t.fontSize(60)
            # add some text
            t += "0123456789 Hello"
            # enable some open type features
            t.openTypeFeatures(smcp=True, lnum=True)
            # add some text
            t += " 0123456789 Hello"
            # draw the formatted string
            text(t, (10, 80))


        NOTE: signature differs from DrawBot:

        ``def openTypeFeatures(self, *args, **features):``

        >>> from pagebot import getContext
        >>> context = getContext()
        >>> context.newDrawing()
        >>> context.newPage(420, 420)
        >>> context.openTypeFeatures(dict(smcp=True, zero=True))
        """
        self.b.openTypeFeatures(**features)

    def listOpenTypeFeatures(self, fontName=None):
        """Answers the list of opentype features available in the named
        font."""
        return self.b.listOpenTypeFeatures(fontName)

    def fontVariations(self, *args, **axes):
        return self.b.fontVariations(*args, **axes)

    def listFontVariations(self, fontName=None):
        return self.b.listFontVariations(fontName=fontName)

    # TODO
    # Future experiment, making UI/Vanilla layout for apps by PageBot
    # Needs some additional conceptual thinking.

    #   U I  components based on Vanilla API
    def window(self, title=None, x=None, y=None, w=None, h=None, style=None,
        minW=None, maxW=None, minH=None, maxH=None, closable=None, **kwargs):
        """Create and opening a window, using Vanilla.
        """
        """
        FIXME
        >>> from pagebot import getContext
        >>> context = getContext('DrawBot')
        >>> from pagebot.toolbox.units import pt, mm
        >>> window = context.window('My Window', 50, 50, pt(200), mm(50))
        >>> window.open()
        """

        '''
        if x is None:
            x = DEFAULT_WINX
        if y is None:
            y = DEFAULT_WINY
        if w is None:
            w = DEFAULT_WINW
        if h is None:
            h = DEFAULT_WINH
        posSize = upt(x), upt(y), upt(w), upt(h)
        if minW is None and minH is None:
            minSize = None
        else:
            minSize = minW or w, minH or h
        if maxW is None and maxH is None:
            maxSize = None
        else:
            maxSize = maxW or w, maxH or h
        if closable is None:
            closable = True

        return Window(posSize, title=title or 'Untitled',
            minSize=minSize, maxSize=maxSize, closable=closable)
        '''

    def group(self, x=None, y=None, w=None, h=None, **kwargs):
        #return Group((upt(x) or 0, upt(y) or 0, upt(w) or 0, upt(h) or 0))
        pass

    def button(self, title=None, x=None, y=None, w=None, h=None, style=None,
            callback=None, **kwargs):
        """Create a Vanilla button"""
        #return Button((upt(x) or 0, upt(y) or 0, upt(w) or 0, upt(h) or 0),
        #title or 'Button', callback=callback)

    def canvas(self, x=None, y=None, w=None, h=None):
        """Answer an instance of the DrawBot drawing canvas."""
        #return drawBot.ui.drawView.DrawView((upt(x or 0), upt(y or 0), upt(w or 0), upt(h or 0)))

    #   S C R E E N

    def screenSize(self):
        """Answers the current screen size in DrawBot. Otherwise default is to
        do nothing. PageBot function.

        >>> from pagebot import getContext
        >>> context = getContext()
        >>> size = context.screenSize()
        >>> size[0] > 100 and size[1] > 100
        True
        """
        return pt(self.b.sizes().get('screen', None))

if __name__ == '__main__':
    import doctest
    import sys
    sys.exit(doctest.testmod()[0])



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# -----------------------------------------------------------------------------
#
#     P A G E B O T
#
#     Copyright (c) 2016+ Buro Petr van Blokland + Claudia Mens
#     www.pagebot.io
#     Licensed under MIT conditions
#
#     Supporting DrawBot, www.drawbot.com
#     Supporting Flat, xxyxyz.org/flat
# -----------------------------------------------------------------------------
#
#     beziercontour.py
#

from fontTools.pens.areaPen import AreaPen

class BezierContour(list):
    """A Bézier contour object."""

    def __init__(self, *args, **kwargs):
        super(BezierContour, self).__init__(*args, **kwargs)
        self.open = True

    def __repr__(self):
        return "<BezierContour>"

    def _get_clockwise(self):
        pen = AreaPen()
        pen.endPath = pen.closePath
        self.drawToPen(pen)
        return pen.value < 0

    clockwise = property(_get_clockwise, doc="A boolean representing if the contour has a clockwise direction.")

    def drawToPointPen(self, pointPen):
        pointPen.beginPath()
        for i, segment in enumerate(self):
            if len(segment) == 1:
                segmentType = "line"
                if i == 0 and self.open:
                    segmentType = "move"
                pointPen.addPoint(segment[0], segmentType=segmentType)
            else:
                pointPen.addPoint(segment[0])
                pointPen.addPoint(segment[1])
                pointPen.addPoint(segment[2], segmentType="curve")
        pointPen.endPath()

    def drawToPen(self, pen):
        for i, segment in enumerate(self):
            if i == 0:
                pen.moveTo(*segment)
            elif len(segment) == 1:
                pen.lineTo(*segment)
            else:
                pen.curveTo(*segment)
        if self.open:
            pen.endPath()
        else:
            pen.closePath()

    def _get_points(self):
        return [point for segment in self for point in segment]

    points = property(_get_points, doc="Return a list of all the points making up this contour, regardless of whether they are on curve or off curve.")




'''
def getTextLines(txt, box):
    """Answers a list of (x,y) positions of all line starts in the box. This
    function may become part of standard DrawBot in the near future."""
    x, y, w, h = box
    attrString = txt.getNSObject()
    setter = CoreText.CTFramesetterCreateWithAttributedString(attrString)
    path = Quartz.CGPathCreateMutable()
    Quartz.CGPathAddRect(path, None, Quartz.CGRectMake(x, y, w, h))
    box = CoreText.CTFramesetterCreateFrame(setter, (0, 0), path, None)
    ctLines = CoreText.CTFrameGetLines(box)
    return ctLines

def getBaselines(txt, box):
    """Answers a list of (x,y) positions of all line starts in the box. This
    function may become part of standard DrawBot in the near future."""
    x, y, w, h = box
    attrString = txt.getNSObject()
    setter = CoreText.CTFramesetterCreateWithAttributedString(attrString)
    path = Quartz.CGPathCreateMutable()
    Quartz.CGPathAddRect(path, None, Quartz.CGRectMake(x, y, w, h))
    box = CoreText.CTFramesetterCreateFrame(setter, (0, 0), path, None)
    ctLines = CoreText.CTFrameGetLines(box)
    #print(ctLines)
    origins = CoreText.CTFrameGetLineOrigins(box, (0, len(ctLines)), None)
    #print(origins)
    return [(x + o.x, y + o.y) for o in origins]

def getTextPositionSearch(bs, w, h, search, xTextAlign=LEFT, hyphenation=True):
    from AppKit import NSLocationInRange
    bc = BaseContext()
    path = CoreText.CGPathCreateMutable()
    CoreText.CGPathAddRect(path, None, CoreText.CGRectMake(0, 0, w, h))

    attrString = bc.attributedString(bs, align=xTextAlign)
    if hyphenation and bc._state.hyphenation:
        attrString = bc.hyphenateAttributedString(attrString, w)

    txt = attrString.string()
    searchRE = re.compile(search)
    locations = []
    for found in searchRE.finditer(txt):
        locations.append((found.start(), found.end()))

    setter = CTFramesetterCreateWithAttributedString(attrString)
    box = CTFramesetterCreateFrame(setter, (0, 0), path, None)

    ctLines = CTFrameGetLines(box)
    origins = CTFrameGetLineOrigins(box, (0, len(ctLines)), None)

    rectangles = []
    for startLocation, endLocation in locations:
        minx = miny = maxx = maxy = None
        for i, (originX, originY) in enumerate(origins):
            ctLine = ctLines[i]
            bounds = CTLineGetImageBounds(ctLine, None)
            if bounds.size.width == 0:
                continue
            _, ascent, descent, leading = CTLineGetTypographicBounds(ctLine, None, None, None)
            height = ascent + descent
            lineRange = CTLineGetStringRange(ctLine)
            miny = maxy = originY

            if NSLocationInRange(startLocation, lineRange):
                minx, _ = CTLineGetOffsetForStringIndex(ctLine, startLocation, None)

            if NSLocationInRange(endLocation, lineRange):
                maxx, _ = CTLineGetOffsetForStringIndex(ctLine, endLocation, None)
                rectangles.append((ctLine, (minx, miny - descent, maxx - minx, height)))

            if minx and maxx is None:
                rectangles.append((ctLine, (minx, miny - descent, bounds.size.width - minx, height)))
                minx = 0

    return rectangles
'''



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# -----------------------------------------------------------------------------
#
#     P A G E B O T
#
#     Copyright (c) 2016+ Buro Petr van Blokland + Claudia Mens
#     www.pagebot.io
#     Licensed under MIT conditions
#
#     Supporting DrawBot, www.drawbot.com
#     Supporting Flat, xxyxyz.org/flat
# -----------------------------------------------------------------------------
#
#     osxcolor.py
#

import AppKit
from pagebotosx.errors import PageBotOSXError

class OSXColor:

    colorSpace = AppKit.NSColorSpace.genericRGBColorSpace

    def __init__(self, r=None, g=None, b=None, a=1):
        self._color = None
        if r is None:
            return

        self.r = r
        self.g = g
        self.b = b
        self.a = a

        if isinstance(r, AppKit.NSColor):
            self._color = r
        elif g is None and b is None:
            self._color = AppKit.NSColor.colorWithCalibratedRed_green_blue_alpha_(r, r, r, a)
        elif b is None:
            self._color = AppKit.NSColor.colorWithCalibratedRed_green_blue_alpha_(r, r, r, g)
        else:
            self._color = AppKit.NSColor.colorWithCalibratedRed_green_blue_alpha_(r, g, b, a)

        self._color = self._color.colorUsingColorSpace_(self.colorSpace())
        super().__init__()

    def set(self):
        self._color.set()

    def setStroke(self):
        self._color.setStroke()

    def getNSObject(self):
        return self._color

    def copy(self):
        new = self.__class__()
        new._color = self._color.copy()
        return new

    @classmethod
    def getColorsFromList(cls, inputColors):
        outputColors = []
        for color in inputColors:
            color = cls.getColor(color)
            outputColors.append(color)
        return outputColors

    @classmethod
    def getColor(cls, color):
        if isinstance(color, cls.__class__):
            return color
        elif isinstance(color, (tuple, list)):
            return cls(*color)
        elif isinstance(color, AppKit.NSColor):
            return cls(color)
        raise PageBotOSXError("Not a valid color: %s" % color)

class OSXCMYKColor(OSXColor):

    colorSpace = AppKit.NSColorSpace.genericCMYKColorSpace

    def __init__(self, c=None, m=None, y=None, k=None, a=1):
        if c is None:
            return
        if isinstance(c, AppKit.NSColor):
            self._color = c
        else:
            self._color = AppKit.NSColor.colorWithDeviceCyan_magenta_yellow_black_alpha_(c, m, y, k, a)
        self._color = self._color.colorUsingColorSpace_(self.colorSpace())
        self._cmyka = c, m, y, k, a
        super().__init__()

class OSXShadow:

    _colorClass = OSXColor

    def __init__(self, offset=None, blur=None, color=None):
        if offset is None:
            return
        self.offset = offset
        self.blur = blur
        self.color = self._colorClass.getColor(color)
        self.cmykColor = None

    def copy(self):
        new = self.__class__()
        new.offset = self.offset
        new.blur = self.blur
        new.color = self.color.copy()
        new.cmykColor = None
        if self.cmykColor:
            new.cmykColor = self.cmykColor.copy()
        return new

class OSXGradient:

    _colorClass = OSXColor

    def __init__(self, gradientType=None, start=None, end=None, colors=None, positions=None, startRadius=None, endRadius=None):
        if gradientType is None:
            return
        if gradientType not in ("linear", "radial"):
            raise PageBotOSXError("Gradient type must be either 'linear' or 'radial'")
        if not colors or len(colors) < 2:
            raise PageBotOSXError("Gradient needs at least 2 colors")
        if positions is None:
            positions = [i / float(len(colors) - 1) for i in range(len(colors))]
        if len(colors) != len(positions):
            raise PageBotOSXError("Gradient needs a correct position for each color")

        self.gradientType = gradientType
        self.colors = self._colorClass.getColorsFromList(colors)
        self.cmykColors = None
        self.positions = positions
        self.start = start
        self.end = end
        self.startRadius = startRadius
        self.endRadius = endRadius

    def copy(self):
        new = self.__class__()
        new.gradientType = self.gradientType
        new.colors = [color.copy() for color in self.colors]
        new.cmykColors = None
        if self.cmykColors:
            new.cmykColors = [color.copy() for color in self.cmykColors]
        new.positions = list(self.positions)
        new.start = self.start
        new.end = self.end
        new.startRadius = self.startRadius
        new.endRadius = self.endRadius
        return new



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# -----------------------------------------------------------------------------
#
#     P A G E B O T
#
#     Copyright (c) 2016+ Buro Petr van Blokland + Claudia Mens
#     www.pagebot.io
#     Licensed under MIT conditions
#
#     Supporting DrawBot, www.drawbot.com
#     Supporting Flat, xxyxyz.org/flat
# -----------------------------------------------------------------------------
#
#     variations.py
#

from copy import copy
from pagebot.fonttoolbox.objects.font import getFont, getInstance
from pagebot.style import css
from pagebot.toolbox.units import upt
from pagebotosx.contexts.drawbot.drawbotstring import pixelBounds

class Variations:

    @classmethod
    def _newFitWidthString(cls, fs, context, fontSize, w, pixelFit):
        if pixelFit:
            tx, _, tw, _ = pixelBounds(fs)
        else:
            tx, tw = 0, context.b.textSize(fs)[0]
        fspt, wpt, txpt, twpt = upt(fontSize, w, tx, tw)

        # Render the units, to compare for zero division.
        dxpt = twpt - txpt
        if dxpt:
            return wpt * fspt / dxpt

        # Zero division, cannot calculate.
        return None

    @classmethod
    def _newFitHeightString(cls, fs, context, fontSize, h, pixelFit):
        if pixelFit:
            _, ty, _, th = pixelBounds(fs)
        else:
            ty, th = 0, context.b.textSize(fs)[1]

        fspt, hpt, typt, thpt = upt(fontSize, h, ty, th)

        # Render the units, to compare for zero division.
        dypt = thpt - typt
        if dypt:
            return hpt * fspt / dypt

        # Zero division, cannot calculate.
        return None

    FITTING_TOLERANCE = 3

    @classmethod
    def fitString(cls, t, context, e=None, style=None, w=None, h=None,
            useXTRA=True, pixelFit=True):
        """Answers the DrawBotString instance from valid attributes in style.
        Set all values after testing their existence, so they can inherit from
        previous style formats in the string. If the target width w and height
        are defined, and if there is a [wdth] or [XTRA] axis in the current
        Variable Font, then values are iterated to make the best location /
        instance for the rectangle fit. In case the fontSize is set and the
        width w is set, then just use the [wdth] or [XTRA] to make a horizontal
        fit, keeping the size. If the axes run to extreme, the string is return
        without changing width. In case a font path was supplied, then try
        to get a Font instance for that path, as we need to test it for
        existing axes as Variable Font.

        FIXME: Fitting does not work anymore.
        TODO: Move to base.

        >>> from pagebot import getContext
        >>> from pagebot.toolbox.color import blackColor
        >>> context = getContext('DrawBot')
        >>> from pagebot.fonttoolbox.objects.font import findFont
        >>> font = findFont('RobotoDelta-VF')
        >>> #font = findFont('Fit-Variable_1') # DJR-Fit needs to be installed.
        >>> style = dict(font=font, textFill=blackColor, textStroke=noColor)

        >>> 'wdth' in font.axes.keys() or 'XTRA' in font.axes.keys() # One of them is there
        True

        """

        """
        >>> s = DrawBotString.newString('Hello', context, style=style, w=pt(300))
        >>> s.bounds() # Rounded width
        (297, 195)
        >>> s = DrawBotString.fitString('Hello', context, style=style, w=pt(400), h=pt(220))
        >>> int(round(s.bounds()[2]-s.bounds()[0])) # Rounded pixel width
        399
        >>> int(round(s.bounds()[3]-s.bounds()[1])) # Rounded pixel height
        220
        >>> #s.bounds()

        """
        style = copy(style)

        # In case the used already supplied a VF location, use it.
        location = copy(css('', e, style, default={}))
        font = css('font', e, style)

        # Assuming it's a path, get the Font instance.
        if isinstance(font, str):
            # Make sure we gave a real Font instance.
            font = getFont(font)
        style['font'] = font

        # Get the flag if fit locations should be rounded (less cached
        # instance) or accurate.
        roundVariableFitLocation = style.get('roundVariableFitLocation', True)

        # In case font is not a variable font, or not [wdth] or [XTRA] present,
        # then using normal string fit is the best we can do.
        if not 'wdth' in font.axes and not 'XTRA' in font.axes:
            return cls.newString(t, context, e=e, style=style, w=w, h=h, pixelFit=pixelFit)

        # Decide which axis to use for width adjustments and get the axis
        # values.
        if not useXTRA or not 'XTRA' in font.axes:
            # Try to force usage of [XTRA] if it exists, otherwise use[wdth]
            axisTag = 'wdth'
        else:
            axisTag = 'XTRA'
        minValue, defaultValue, maxValue = font.axes[axisTag]

        if h is not None:
            # Fitting in heigt, calculate/iterate for the fitting font size.
            bs = cls.newString(t, context, e=e, style=style, h=h, pixelFit=pixelFit)
            style['fontSize'] = bs.fittingFontSize
        else:
            # Assuming there is a fontSize set, we'll use that as vertical
            # requirement.
            bs = cls.newString(t, context, e=e, style=style, pixelFit=pixelFit)

        # Now we have a formatted string with a given fontSize, guess to fit on
        # the width.

        # Get pixel bounds of the string
        tx, _, tw, _ = bs.bounds()

        # Pixel width of the current string.
        tw = tw - tx

        # Testing if something changed, for extreme of axes.
        prevW = None
        axisValue = defaultValue


        # Limit the maximum amount of iterations as safeguard.
        for n in range(100):
            # Too wide, try iterate smaller in ratio of wdth / XTRA axis
            # values.
            if tw > w:
                # Clip wide range to current.
                maxValue = axisValue
                # Guess the new axisvalue from the ratio of tw / w
                axisValue = (axisValue - minValue)/2 + minValue
                if roundVariableFitLocation:
                    axisValue = int(round(axisValue))
                loc = copy(location)
                loc[axisTag] = axisValue
                loc['opsz'] = upt(style['fontSize'])
                style['font'] = getInstance(font, loc)
                bs = cls.newString(t, context, e=e, style=style, pixelFit=pixelFit)

                # Get pixel bounds of the string.
                tx, ty, tw, th = bs.bounds()

                # Total width for the current.
                tw = tw - tx

                # Did not change, probably not able to get more condensed.
                if prevW == tw:
                    break
                prevW = tw

            # Too condensed, try to make wider.
            elif tw < w - cls.FITTING_TOLERANCE:
                # Clip narrow range to current
                minValue = axisValue
                axisValue = (maxValue - axisValue)/2 + axisValue
                if roundVariableFitLocation:
                    axisValue = int(round(axisValue))
                loc = copy(location)
                loc[axisTag] = axisValue
                loc['opsz'] = upt(style['fontSize'])
                style['font'] = getInstance(font, loc)
                bs = cls.newString(t, context, e=e, style=style, pixelFit=pixelFit)

                # Get pixel bounds of the string.
                tx, ty, tw, th = bs.bounds()

                # Total width for the current.
                tw = tw - tx

                # Did not change, probably not able to get more condensed.
                if prevW == tw:
                    break

                prevW = tw


            # We found a fitting VF-location within tolerance. Back out.
            else:
                break
        return bs

    """
    # FIXME: Disabled string fitting here. Use fitString(...) instead.
    if False and w is not None:
        # A target width is already defined, calculate again with the
        # fontSize ratio correction. We use the enclosing pixel bounds
        # instead of the context.textSide(newT) here, because it is more
        # consistent for tracked text. context.textSize will add space to
        # the right of the string.
        attrs = copy(attrs)
        attrs['textFill'] = attrs.get('fill')
        attrs['textStroke'] = attrs.get('stroke')
        attrs['textStrokeWidth'] = attrs.get('strokeWidth')

        fittingFontSize = cls._newFitWidthString(newT, context, attrs.get('fontSize', DEFAULT_FONT_SIZE), w, pixelFit)
        if fittingFontSize is not None: # Checked on zero division
            # Repair the attrs to style, so it can be reused for new string
            attrs['fontSize'] = fittingFontSize
            newS = cls.newString(t, context, style=attrs)
            # Test the width we got by linear interpolation. Scale back if still too large.
            # Iterate until it really fits.
            while newS.size[0] > w and attrs['fontSize']:
                attrs['fontSize'] -= 0.1 # Incremental decrease the size until it fits
                newS = cls.newString(t, context, style=attrs)
        else:
            newS = cls(newT, context, attrs) # Cannot fit, answer untouched.
            isFitting = False
    elif False and h is not None:
        # A target height is already defined, calculate again with the
        # fontSize ratio correction. We use the enclosing pixel bounds
        # instead of the context.textSide(newT) here, because it is
        # more consistent for tracked text. context.textSize will add space
        # to the right of the string.
        attrs = copy(attrs)
        attrs['fontSize'] = fittingFontSize
        attrs['textFill'] = attrs.get('fill')
        attrs['textStroke'] = attrs.get('stroke')
        attrs['textStrokeWidth'] = attrs.get('strokeWidth')

        fittingFontSize = cls._newFitHeightString(newT, context, attrs.get('fontSize', DEFAULT_FONT_SIZE), h, pixelFit)

        if fittingFontSize is not None:
            # Repair the attrs to style, so it can be reused for new string
            newS = cls.newString(t, context, style=attrs)
            didFit = True
        else:
            newS = cls(newT, context, attrs) # Cannot fit, answer untouched.
            isFitting = False
    else:
    """




#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# -----------------------------------------------------------------------------
#
#     P A G E B O T
#
#     Copyright (c) 2016+ Buro Petr van Blokland + Claudia Mens
#     www.pagebot.io
#     Licensed under MIT conditions
#
#     Supporting DrawBot, www.drawbot.com
#     Supporting Flat, xxyxyz.org/flat
# -----------------------------------------------------------------------------
#
#     transform.py
#

from fontTools.misc.transform import Transform

def transformationAtCenter(matrix, centerPoint):
    """Helper function for rotate(), scale() and skew() to apply a
    transformation with a specified center point.

        >>> transformationAtCenter((2, 0, 0, 2, 0, 0), (0, 0))
        (2, 0, 0, 2, 0, 0)
        >>> transformationAtCenter((2, 0, 0, 2, 0, 0), (100, 200))
        (2, 0, 0, 2, -100, -200)
        >>> transformationAtCenter((-2, 0, 0, 2, 0, 0), (100, 200))
        (-2, 0, 0, 2, 300, -200)
        >>> t = Transform(*transformationAtCenter((0, 1, 1, 0, 0, 0), (100, 200)))
        >>> t.transformPoint((100, 200))
        (100, 200)
        >>> t.transformPoint((0, 0))
        (-100, 100)
    """
    if centerPoint == (0, 0):
        return matrix

    t = Transform()
    cx, cy = centerPoint
    t = t.translate(cx, cy)
    t = t.transform(matrix)
    t = t.translate(-cx, -cy)
    return tuple(t)





#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# -----------------------------------------------------------------------------
#
#     P A G E B O T
#
#     Copyright (c) 2016+ Buro Petr van Blokland + Claudia Mens
#     www.pagebot.io
#     Licensed under MIT conditions
#
#     Supporting DrawBot, www.drawbot.com
#     Supporting Flat, xxyxyz.org/flat
# -----------------------------------------------------------------------------
#
#     formattedstring.py
#

import os
import logging
import AppKit
import CoreText

from fontTools.ttLib import TTFont, TTLibError
from fontTools.misc.macRes import ResourceReader, ResourceError

# FIXME: featureMap deprecated, use CTFeatureOpenTypeTag?
# https://github.com/typemytype/drawbot/pull/307
from drawBot.context.tools.openType import getFeatureTagsForFontName#, featureMap
from drawBot.context.tools.variation import *
from pagebotosx.errors import PageBotOSXError
from pagebotosx.osxcolor import OSXColor, OSXCMYKColor
from pagebotosx.strings.tryinstallfont import _tryInstallFontFromFontName

logger = logging.getLogger(__name__)

_FALLBACKFONT = "LucidaGrande"

class FormattedString:
    """Based on drawBot.context.baseContext.FormattedString.

    FormattedString is a reusable object, if you want to draw the same string
    multiple times. FormattedString objects can be drawn with the `text(txt,
    (x, y))` and `textBox(txt, (x, y, w, h))` methods."""

    _colorClass = OSXColor
    _cmykColorClass = OSXCMYKColor

    _textAlignMap = dict(
        center=AppKit.NSCenterTextAlignment,
        left=AppKit.NSLeftTextAlignment,
        right=AppKit.NSRightTextAlignment,
        justified=AppKit.NSJustifiedTextAlignment,
    )

    _textTabAlignMap = dict(
        center=AppKit.NSCenterTextAlignment,
        left=AppKit.NSLeftTextAlignment,
        right=AppKit.NSRightTextAlignment,
    )

    _textUnderlineMap = dict(
        single=AppKit.NSUnderlineStyleSingle,
        # thick=AppKit.NSUnderlineStyleThick,
        # double=AppKit.NSUnderlineStyleDouble,
        # solid=AppKit.NSUnderlinePatternSolid,
        # dotted=AppKit.NSUnderlinePatternDot,
        # dashed=AppKit.NSUnderlinePatternDash,
        # dotDashed=AppKit.NSUnderlinePatternDashDot,
        # dotDotted=AppKit.NSUnderlinePatternDashDotDot,
        # byWord=0x8000 # AppKit.NSUnderlineByWord,
    )

    _formattedAttributes = dict(
        font=_FALLBACKFONT,
        fallbackFont=None,
        fontSize=10,

        fill=(0, 0, 0),
        cmykFill=None,
        stroke=None,
        cmykStroke=None,
        strokeWidth=1,

        align=None,
        lineHeight=None,
        tracking=None,
        baselineShift=None,
        underline=None,
        openTypeFeatures=dict(),
        fontVariations=dict(),
        tabs=None,
        indent=None,
        tailIndent=None,
        firstLineIndent=None,
        paragraphTopSpacing=None,
        paragraphBottomSpacing=None,

        language=None,
    )

    def __init__(self, txt=None, **kwargs):
        self.clear()
        # create all _<attributes> in the formatted text object
        # with default values
        for key, value in self._formattedAttributes.items():
            if isinstance(value, dict):
                value = dict(value)
            if isinstance(value, list):
                value = list(value)
            setattr(self, "_%s" % key, value)
        attributes = self._validateAttributes(kwargs, addDefaults=False)
        if txt:
            self.append(txt, **attributes)
        else:
            # call each method with the provided value
            for key, value in attributes.items():
                self._setAttribute(key, value)
            self._setColorAttributes(attributes)

    def _setAttribute(self, attribute, value):
        method = getattr(self, attribute)
        if isinstance(value, (list, tuple)):
            method(*value)
        elif isinstance(value, dict):
            method(**value)
        else:
            method(value)

    def _setColorAttributes(self, attributes):
        colorAttributeNames = ("fill", "stroke", "cmykFill", "cmykStroke")
        for key in colorAttributeNames:
            value = attributes.get(key)
            if value is not None:
                setattr(self, "_%s" % key, value)

        if self._fill is not None:
            try:
                len(self._fill)
            except Exception:
                self._fill = (self._fill,)
        if self._stroke is not None:
            try:
                len(self._stroke)
            except Exception:
                self._stroke = (self._stroke,)
        if self._fill:
            self._cmykFill = None
        elif self._cmykFill:
            self._fill = None

        if self._stroke:
            self._cmykStroke = None
        elif self._cmykStroke:
            self._stroke = None

    def _validateAttributes(self, attributes, addDefaults=True):
        for attribute in attributes:
            if attribute not in self._formattedAttributes:
                raise TypeError("FormattedString got an unexpected keyword argument '%s'" % attribute)
        result = dict()
        if addDefaults:
            for key, value in self._formattedAttributes.items():
                if isinstance(value, dict):
                    value = dict(value)
                elif isinstance(value, list):
                    value = list(value)
                result[key] = value
        result.update(attributes)
        return result

    def clear(self):
        self._attributedString = AppKit.NSMutableAttributedString.alloc().init()
        self._openTypeFeatures = None

    def append(self, txt, **kwargs):
        """Adds `txt` to the formatted string with some additional text
        formatting attributes:

        * `font`: the font to be used for the given text, if a font path is
          given the font will be installed and used directly.
        * `fallbackFont`: the fallback font
        * `fontSize`: the font size to be used for the given text
        * `fill`: the fill color to be used for the given text
        * `cmykFill`: the cmyk fill color to be used for the given text
        * `stroke`: the stroke color to be used for the given text
        * `cmykStroke`: the cmyk stroke color to be used for the given text
        * `strokeWidth`: the strokeWidth to be used for the given text
        * `align`: the alignment to be used for the given text
        * `lineHeight`: the lineHeight to be used for the given text
        * `tracking`: set tracking for the given text
        * `baselineShift`: set base line shift for the given text
        * `openTypeFeatures`: enable OpenType features
        * `fontVariations`: pick a variation by axes values
        * `tabs`: enable tabs
        * `indent`: the indent of a paragraph
        * `tailIndent`: the tail indent of a paragraph
        * `firstLineIndent`: the first line indent of a paragraph
        * `paragraphTopSpacing`: the spacing at the top of a paragraph
        * `paragraphBottomSpacing`: the spacing at the bottom of a paragraph
        * `language`: the language of the text

        All formatting attributes follow the same notation as other similar
        DrawBot methods.  A color is a tuple of `(r, g, b, alpha)`, and a
        cmykColor is a tuple of `(c, m, y, k, alpha)`.

        Text can also be added with `formattedString += "hello"`. It will
        append the text with the current settings of the formatted string.
        """
        attributes = self._validateAttributes(kwargs, addDefaults=False)

        for key, value in attributes.items():
            self._setAttribute(key, value)

        self._setColorAttributes(attributes)

        if isinstance(txt, FormattedString):
            self._attributedString.appendAttributedString_(txt.getNSObject())
            return

        attributes = {}

        if self._font:
            font = AppKit.NSFont.fontWithName_size_(self._font, self._fontSize)
            if font is None:
                ff = self._fallbackFont
                if ff is None:
                    ff = _FALLBACKFONT
                logger.warning("font: '%s' is not installed, back to the fallback font: '%s'", self._font, ff)
                font = AppKit.NSFont.fontWithName_size_(ff, self._fontSize)
            coreTextfeatures = []

            if self._openTypeFeatures:
                existingOpenTypeFeatures = getFeatureTagsForFontName(self._font)
                # sort features by their on/off state
                # set all disabled features first
                orderedOpenTypeFeatures = sorted(self._openTypeFeatures.items(), key=lambda kv: kv[1])
                for featureTag, value in orderedOpenTypeFeatures:
                    coreTextFeatureTag = featureTag
                    if not value:
                        coreTextFeatureTag = "%s_off" % featureTag


                    '''
                    if coreTextFeatureTag in featureMap:
                        if value and featureTag not in existingOpenTypeFeatures:
                            # only warn when the feature is on and not existing for the current font
                            logger.warning("OpenType feature '%s' not available for '%s'", featureTag, self._font)
                        feature = featureMap[coreTextFeatureTag]
                        coreTextfeatures.append(feature)
                    else:
                        logger.warning("OpenType feature '%s' not available", featureTag)
                    '''

            coreTextFontVariations = dict()

            if self._fontVariations:
                existingAxes = getVariationAxesForFontName(self._font)
                for axis, value in self._fontVariations.items():
                    if axis in existingAxes:
                        existinsAxis = existingAxes[axis]
                        # clip variation value within the min max value
                        if value < existinsAxis["minValue"]:
                            value = existinsAxis["minValue"]
                        if value > existinsAxis["maxValue"]:
                            value = existinsAxis["maxValue"]
                        coreTextFontVariations[convertVariationTagToInt(axis)] = value
                    else:
                        logger.warning("variation axis '%s' not available for '%s'", axis, self._font)
            fontAttributes = {}
            if coreTextfeatures:
                fontAttributes[CoreText.NSFontFeatureSettingsAttribute] = coreTextfeatures
            if coreTextFontVariations:
                fontAttributes[CoreText.NSFontVariationAttribute] = coreTextFontVariations
            if self._fallbackFont:
                fontAttributes[CoreText.NSFontCascadeListAttribute] = [AppKit.NSFontDescriptor.fontDescriptorWithName_size_(self._fallbackFont, self._fontSize)]
            fontDescriptor = font.fontDescriptor()
            fontDescriptor = fontDescriptor.fontDescriptorByAddingAttributes_(fontAttributes)
            font = AppKit.NSFont.fontWithDescriptor_size_(fontDescriptor, self._fontSize)
            attributes[AppKit.NSFontAttributeName] = font
        elif self._fontSize:
            font = AppKit.NSFont.fontWithName_size_(_FALLBACKFONT, self._fontSize)
            attributes[AppKit.NSFontAttributeName] = font
        if self._fill or self._cmykFill:
            if self._fill:
                fillColor = self._colorClass.getColor(self._fill).getNSObject()
            elif self._cmykFill:
                fillColor = self._cmykColorClass.getColor(self._cmykFill).getNSObject()
            attributes[AppKit.NSForegroundColorAttributeName] = fillColor
        else:
            # Seems like the default foreground color is black. Set clear
            # color when the fill is None.
            attributes[AppKit.NSForegroundColorAttributeName] = AppKit.NSColor.clearColor()

        if self._stroke or self._cmykStroke:
            if self._stroke:
                strokeColor = self._colorClass.getColor(self._stroke).getNSObject()
            elif self._cmykStroke:
                strokeColor = self._cmykColorClass.getColor(self._cmykStroke).getNSObject()

            attributes[AppKit.NSStrokeColorAttributeName] = strokeColor

            # stroke width must be negative
            # Supply a negative value for NSStrokeWidthAttributeName
            # when you wish to draw a string that is both filled and stroked.
            # see https://developer.apple.com/library/content/qa/qa1531/_index.html
            attributes[AppKit.NSStrokeWidthAttributeName] = -abs(self._strokeWidth)

        para = AppKit.NSMutableParagraphStyle.alloc().init()

        if self._align:
            para.setAlignment_(self._textAlignMap[self._align])

        if self._tabs:
            for tabStop in para.tabStops():
                para.removeTabStop_(tabStop)

            if len(self._tabs) < 12:
                self._tabs = list(self._tabs)
                # add tab stops if there is not enough stops...
                # the default is 12 tabs, so lets add 12 in steps of 28
                lastTabValue = self._tabs[-1][0]
                for tabIndex in range(12 - len(self._tabs)):
                    self._tabs.append((lastTabValue + 28 * (tabIndex + 1), "left"))

            for tab, tabAlign in self._tabs:
                tabOptions = None
                if tabAlign in self._textTabAlignMap:
                    tabAlign = self._textTabAlignMap[tabAlign]
                else:
                    tabCharSet = AppKit.NSCharacterSet.characterSetWithCharactersInString_(tabAlign)
                    tabOptions = {AppKit.NSTabColumnTerminatorsAttributeName: tabCharSet}
                    tabAlign = self._textAlignMap["right"]
                tabStop = AppKit.NSTextTab.alloc().initWithTextAlignment_location_options_(tabAlign, tab, tabOptions)
                para.addTabStop_(tabStop)
        if self._lineHeight is not None:
            # para.setLineSpacing_(0.0)
            # para.setLineHeightMultiple_(1)
            para.setMinimumLineHeight_(self._lineHeight)
            para.setMaximumLineHeight_(self._lineHeight)

        if self._indent is not None:
            para.setHeadIndent_(self._indent)
            para.setFirstLineHeadIndent_(self._indent)
        if self._tailIndent is not None:
            para.setTailIndent_(self._tailIndent)
        if self._firstLineIndent is not None:
            para.setFirstLineHeadIndent_(self._firstLineIndent)

        if self._paragraphTopSpacing is not None:
            para.setParagraphSpacingBefore_(self._paragraphTopSpacing)
        if self._paragraphBottomSpacing is not None:
            para.setParagraphSpacing_(self._paragraphBottomSpacing)

        if self._tracking:
            attributes[AppKit.NSKernAttributeName] = self._tracking
        if self._baselineShift is not None:
            attributes[AppKit.NSBaselineOffsetAttributeName] = self._baselineShift
        if self._underline in self._textUnderlineMap:
            attributes[AppKit.NSUnderlineStyleAttributeName] = self._textUnderlineMap[self._underline]
        if self._language:
            attributes["NSLanguage"] = self._language
        attributes[AppKit.NSParagraphStyleAttributeName] = para
        txt = AppKit.NSAttributedString.alloc().initWithString_attributes_(txt, attributes)
        self._attributedString.appendAttributedString_(txt)

    def __add__(self, txt):
        new = self.copy()
        if isinstance(txt, self.__class__):
            new.getNSObject().appendAttributedString_(txt.getNSObject())
        else:
            if not isinstance(txt, str):
                raise TypeError("FormattedString requires a string, got '%s'" % type(txt))
            new.append(txt)
        return new

    def __getitem__(self, index):
        if isinstance(index, slice):
            start = index.start
            stop = index.stop
            textLength = len(self)

            if start is None:
                start = 0
            elif start < 0:
                start = textLength + start
            elif start > textLength:
                start = textLength

            if stop is None:
                stop = textLength
            elif stop < 0:
                stop = textLength + stop

            if start + (stop - start) > textLength:
                stop = textLength

            location = start
            length = stop - start

            if location < 0:
                location = 0
            if length > textLength:
                length = textLength
            elif length < 0:
                length = 0

            rng = location, length
            attributes = {key: getattr(self, "_%s" % key) for key in self._formattedAttributes}
            new = self.__class__(**attributes)
            try:
                new._attributedString = self._attributedString.attributedSubstringFromRange_(rng)
            except Exception:
                pass
            return new
        else:
            text = self._attributedString.string()
            return text[index]

    def __len__(self):
        return self._attributedString.length()

    def __repr__(self):
        return self._attributedString.string()

    def font(self, font, fontSize=None):
        """Sets a font with the name of the font. If a font path is given the
        font will be installed and used directly. Optionally a `fontSize` can
        be set directly. The default font, also used as fallback font, is
        'LucidaGrande'. The default `fontSize` is 10pt.

        The name of the font relates to the font's postscript name.

        The font name is returned, which is handy when the font was loaded
        from a path."""
        font = _tryInstallFontFromFontName(font)
        font = str(font)
        self._font = font
        if fontSize is not None:
            self._fontSize = fontSize
        return font

    def fallbackFont(self, font):
        """Sets a fallback font, used whenever a glyph is not available in the
        normal font. If a font path is given the font will be installed and
        used directly."""
        if font:
            font = _tryInstallFontFromFontName(font)
            font = str(font)
            testFont = AppKit.NSFont.fontWithName_size_(font, self._fontSize)
            if testFont is None:
                raise PageBotOSXError("Fallback font '%s' is not available" % font)
        self._fallbackFont = font
        return font

    def fontSize(self, fontSize):
        """Set the font size in points. The default `fontSize` is 10pt."""
        self._fontSize = fontSize

    def fill(self, *fill):
        """Sets the fill color with a `red`, `green`, `blue` and `alpha` value.
        Each argument must a value float between 0 and 1."""
        if fill and fill[0] is None:
            fill = None
        self._fill = fill
        self._cmykFill = None

    def stroke(self, *stroke):
        """Sets the stroke color with a `red`, `green`, `blue` and `alpha`
        value. Each argument must a value float between 0 and 1."""
        if stroke and stroke[0] is None:
            stroke = None
        self._stroke = stroke
        self._cmykStroke = None

    def cmykFill(self, *cmykFill):
        """Sets a fill using a CMYK color before drawing a shape. This is handy
        if the file is intended for print.

        Sets the CMYK fill color. Each value must be a float between 0.0 and
        1.0."""
        if cmykFill and cmykFill[0] is None:
            cmykFill = None
        self._cmykFill = cmykFill
        self._fill = None

    def cmykStroke(self, *cmykStroke):
        """Set a stroke using a CMYK color before drawing a shape. This is
        handy if the file is intended for print.

        Sets the CMYK stroke color. Each value must be a float between 0.0 and
        1.0."""
        if cmykStroke and cmykStroke[0] is None:
            cmykStroke = None
        self._cmykStroke = cmykStroke
        self._stroke = None

    def strokeWidth(self, strokeWidth):
        """Sets stroke width."""
        self._strokeWidth = strokeWidth

    def align(self, align):
        """Sets the text alignment. Possible `align` values are: `left`,
        `center` and `right`."""
        self._align = align

    def lineHeight(self, lineHeight):
        """Sets the line height."""
        self._lineHeight = lineHeight

    def tracking(self, tracking):
        """Sets the tracking between characters."""
        self._tracking = tracking

    def baselineShift(self, baselineShift):
        """Sets the shift of the baseline."""
        self._baselineShift = baselineShift

    def underline(self, underline):
        """Set the underline value. Underline must be `single` or `None`."""
        self._underline = underline

    def openTypeFeatures(self, *args, **features):
        """Enables OpenType features and return the current openType features
        settings.

        If no arguments are given `openTypeFeatures()` will just return the
        current openType features settings.

        .. downloadcode:: openTypeFeaturesFormattedString.py

            size(1000, 200)
            # create an empty formatted string object
            t = FormattedString()
            # set a font
            t.font("ACaslonPro-Regular")
            # set a font size
            t.fontSize(60)
            # add some text
            t += "0123456789 Hello"
            # enable some open type features
            t.openTypeFeatures(smcp=True, lnum=True)
            # add some text
            t += " 0123456789 Hello"
            # draw the formatted string
            text(t, (10, 80))
        """
        if args and features:
            raise PageBotOSXError("Can't combine positional arguments and keyword arguments")
        if args:
            if len(args) != 1:
                raise PageBotOSXError("There can only be one positional argument")
            if args[0] is not None:
                raise PageBotOSXError("First positional argument can only be None")
            logger.warning("openTypeFeatures(None) is deprecated, use openTypeFeatures(resetFeatures=True) instead.")
            self._openTypeFeatures.clear()
        else:
            if features.pop("resetFeatures", False):
                self._openTypeFeatures.clear()
            self._openTypeFeatures.update(features)
        currentFeatures = self.listOpenTypeFeatures()
        currentFeatures.update(self._openTypeFeatures)
        return currentFeatures

    def listOpenTypeFeatures(self, fontName=None):
        """Lists all OpenType feature tags for the current font.

        Optionally a `fontName` can be given. If a font path is given the font
        will be installed and used directly."""
        if fontName:
            fontName = _tryInstallFontFromFontName(fontName)
        else:
            fontName = self._font
        return getFeatureTagsForFontName(fontName)

    def fontVariations(self, *args, **axes):
        """Picks a variation by axes values and return the current font
        variations settings.

        If no arguments are given `fontVariations()` will just return the
        current font variations settings."""
        if args and axes:
            raise PageBotOSXError("Can't combine positional arguments and keyword arguments")

        if args:
            if len(args) != 1:
                raise PageBotOSXError("There can only be one positional argument")
            if args[0] is not None:
                raise PageBotOSXError("First positional argument can only be None")
            logger.warning("fontVariations(None) is deprecated, use fontVariations(resetVariations=True) instead.")
            self._fontVariations.clear()
        else:
            if axes.pop("resetVariations", False):
                self._fontVariations.clear()
            self._fontVariations.update(axes)

        defaultVariations = self.listFontVariations()
        currentVariation = {axis: data["defaultValue"] for axis, data in defaultVariations.items()}
        currentVariation.update(self._fontVariations)
        return currentVariation

    def listFontVariations(self, fontName=None):
        """Lists all variation axes for the current font.

        Optionally a `fontName` can be given. If a font path is given the font
        will be installed and used directly."""
        if fontName:
            fontName = _tryInstallFontFromFontName(fontName)
        else:
            fontName = self._font
        return getVariationAxesForFontName(fontName)

    def tabs(self, *tabs):
        """Sets tabs, tuples of (`float`, `alignment`).  Aligment can be
        `"left"`, `"center"`, `"right"` or any other character. If a character
        is provided the alignment will be `right` and centered on the specified
        character.

        .. downloadcode:: tabsFormattedString.py

            # create a new formatted string
            t = FormattedString()
            # set some tabs
            t.tabs((85, "center"), (232, "right"), (300, "left"))
            # add text with tabs
            t += " hello w o r l d".replace(" ", "\\t")
            # draw the string
            text(t, (10, 10))
        """
        if tabs and tabs[0] is None:
            self._tabs = None
        else:
            self._tabs = tabs

    def indent(self, indent):
        """
        Sets indent of text left of the paragraph.

        .. downloadcode:: indent.py

            # setting up some variables
            x, y, w, h = 10, 10, 500, 600

            txtIndent = 100
            txtFirstLineIndent = 200
            txtTailIndent = -100
            txtFontSize = 22

            paragraphTop = 3
            paragraphBottom = 10

            txt = '''DrawBot is an ideal tool to teach the basics of programming. Students get colorful graphic treats while getting familiar with variables, conditional statements, functions and what have you. Results can be saved in a selection of different file formats, including as high resolution, scaleable PDF, svg, movie, png, jpeg, tiff...'''

            # a new page with preset size
            newPage(w+x*2, h+y*2)
            # draw text indent line
            stroke(1, 0, 0)
            line((x+txtIndent, y), (x+txtIndent, y+h))
            # draw text firstline indent line
            stroke(1, 1, 0)
            line((x+txtFirstLineIndent, y), (x+txtFirstLineIndent, y+h))
            # draw tail indent
            pos = txtTailIndent
            # tail indent could be negative
            if pos <= 0:
                # substract from width of the text box
                pos = w + pos
            stroke(0, 0, 1)
            line((x+pos, y), (x+pos, y+h))
            # draw a rectangle
            fill(0, .1)
            stroke(None)
            rect(x, y, w, h)

            # create a formatted string
            t = FormattedString(fontSize=txtFontSize)
            # set alignment
            t.align("justified")
            # add text
            t += txt
            # add hard return
            t += "\\n"
            # set style for indented text
            t.fontSize(txtFontSize*.6)
            t.paragraphTopSpacing(paragraphTop)
            t.paragraphBottomSpacing(paragraphBottom)
            t.firstLineIndent(txtFirstLineIndent)
            t.indent(txtIndent)
            t.tailIndent(txtTailIndent)
            # add text
            t += txt
            # add hard return
            t += "\\n"
            # reset style
            t.fontSize(txtFontSize)
            t.indent(None)
            t.tailIndent(None)
            t.firstLineIndent(None)
            t.paragraphTopSpacing(None)
            t.paragraphBottomSpacing(None)
            # add text
            t += txt
            # draw formatted string in a text box
            textBox(t, (x, y, w, h))
        """
        self._indent = indent

    def tailIndent(self, indent):
        """Sets indent of text right of the paragraph."""
        self._tailIndent = indent

    def firstLineIndent(self, indent):
        """Sets indent of the text only for the first line."""
        self._firstLineIndent = indent

    def paragraphTopSpacing(self, value):
        """Sets paragraph spacing at the top."""
        self._paragraphTopSpacing = value

    def paragraphBottomSpacing(self, value):
        """Sets paragraph spacing at the bottom."""
        self._paragraphBottomSpacing = value

    def language(self, language):
        """Sets the preferred language as language tag or None to use the
        default language."""
        self._language = language

    def size(self):
        """Returns the size of the text."""
        return self._attributedString.size()

    def getNSObject(self):
        return self._attributedString

    def copy(self):
        """Copies the formatted string."""
        attributes = {key: getattr(self, "_%s" % key) for key in self._formattedAttributes}
        new = self.__class__(**attributes)
        new._attributedString = self._attributedString.mutableCopy()
        return new

    def fontContainsCharacters(self, characters):
        """Returns a Boolean if the current font contains the provided
        `characters`. Characters is a string containing one or more
        characters."""
        font = AppKit.NSFont.fontWithName_size_(self._font, self._fontSize)
        if font is None:
            return False
        result, glyphs = CoreText.CTFontGetGlyphsForCharacters(font, characters, None, len(characters))
        return result

    def fontContainsGlyph(self, glyphName):
        font = AppKit.NSFont.fontWithName_size_(self._font, self._fontSize)
        if font is None:
            return False
        glyph = font.glyphWithName_(glyphName)
        return bool(glyph)

    def fontFilePath(self):
        """Returns the path to the file of the current font."""
        font = AppKit.NSFont.fontWithName_size_(self._font, self._fontSize)
        if font is not None:
            url = CoreText.CTFontDescriptorCopyAttribute(font.fontDescriptor(), CoreText.kCTFontURLAttribute)
            if url:
                return url.path()
        logger.warning("Cannot find the path to the font '%s'.", self._font)
        return None

    def listFontGlyphNames(self):
        """Returns a list of glyph names supported by the current font."""
        path = self.fontFilePath()
        if path is None:
            return []
        try:
            # load the font with fontTools
            # provide a fontNumber as lots of fonts are .ttc font files.
            # search for the res_name_or_index for .dfont files.
            res_name_or_index = None
            fontNumber = None
            ext = os.path.splitext(path)[-1].lower()
            if ext == ".ttc":
                def _getPSName(source):
                    # get PS name
                    name = source["name"]
                    psName = name.getName(6, 1, 0)
                    if psName is None:
                        psName.getName(6, 3, 1)
                    return psName.toStr()

                ttc = TTFont(path, lazy=True, fontNumber=0)
                numFonts = ttc.reader.numFonts
                foundPSName = False
                for fontNumber in range(numFonts):
                    source = TTFont(path, lazy=True, fontNumber=fontNumber)
                    psName = _getPSName(source)
                    if psName == self._font:
                        foundPSName = True
                        break
                if not foundPSName:
                    # fallback to the first font in the ttc.
                    fontNumber = 0

            elif ext == ".dfont":
                try:
                    reader = ResourceReader(path)
                    names = reader.getNames("sfnt")
                    if self._font in names:
                        res_name_or_index = self._font
                    else:
                        res_name_or_index = names[0]
                except ResourceError:
                    pass
            fontToolsFont = TTFont(path, lazy=True, fontNumber=fontNumber, res_name_or_index=res_name_or_index)
        except TTLibError:
            logger.warning("Cannot read the font file for '%s' at the path '%s'", self._font, path)
            return []
        glyphNames = fontToolsFont.getGlyphOrder()
        fontToolsFont.close()
        # remove .notdef from glyph names
        if ".notdef" in glyphNames:
            glyphNames.remove(".notdef")
        return glyphNames

    def fontAscender(self):
        """Returns the current font ascender, based on the current `font` and
        `fontSize`."""
        font = AppKit.NSFont.fontWithName_size_(self._font, self._fontSize)
        if font is None:
            ff = self._fallbackFont or _FALLBACKFONT
            logger.warning("font: '%s' is not installed, back to the fallback font: '%s'", self._font, ff)
            font = AppKit.NSFont.fontWithName_size_(ff, self._fontSize)
        return font.ascender()

    def fontDescender(self):
        """Returns the current font descender, based on the current `font` and
        `fontSize`."""
        font = AppKit.NSFont.fontWithName_size_(self._font, self._fontSize)
        if font is None:
            ff = self._fallbackFont or _FALLBACKFONT
            logger.warning("font: '%s' is not installed, back to the fallback font: '%s'", self._font, ff)
            font = AppKit.NSFont.fontWithName_size_(ff, self._fontSize)
        return font.descender()

    def fontXHeight(self):
        """Returns the current font x-height, based on the current `font` and
        `fontSize`."""
        font = AppKit.NSFont.fontWithName_size_(self._font, self._fontSize)

        if font is None:
            ff = self._fallbackFont or _FALLBACKFONT
            logger.warning("font: '%s' is not installed, back to the fallback font: '%s'", self._font, ff)
            font = AppKit.NSFont.fontWithName_size_(ff, self._fontSize)
        return font.xHeight()

    def fontCapHeight(self):
        """Returns the current font cap height, based on the current `font` and
        `fontSize`. """
        font = AppKit.NSFont.fontWithName_size_(self._font, self._fontSize)

        if font is None:
            ff = self._fallbackFont or _FALLBACKFONT
            logger.warning("font: '%s' is not installed, back to the fallback font: '%s'", self._font, ff)
            font = AppKit.NSFont.fontWithName_size_(ff, self._fontSize)
        return font.capHeight()

    def fontLeading(self):
        """Returns the current font leading, based on the current `font` and
        `fontSize`."""
        font = AppKit.NSFont.fontWithName_size_(self._font, self._fontSize)

        if font is None:
            ff = self._fallbackFont or _FALLBACKFONT
            logger.warning("font: '%s' is not installed, back to the fallback font: '%s'", self._font, ff)
            font = AppKit.NSFont.fontWithName_size_(ff, self._fontSize)
        return font.leading()

    def fontLineHeight(self):
        """Returns the current line height, based on the current `font` and
        `fontSize`. If a `lineHeight` is set, this value will be returned."""
        if self._lineHeight is not None:
            return self._lineHeight
        font = AppKit.NSFont.fontWithName_size_(self._font, self._fontSize)
        if font is None:
            ff = self._fallbackFont or _FALLBACKFONT
            logger.warning("font: '%s' is not installed, back to the fallback font: '%s'", self._font, ff)
            font = AppKit.NSFont.fontWithName_size_(ff, self._fontSize)
        return font.defaultLineHeightForFont()

    def appendGlyph(self, *glyphNames):
        """Appends a glyph by his glyph name using the current `font`.
        Multiple glyph names are possible.

        .. downloadcode:: appendGlyphFormattedString.py

            size(1000, 400)
            # create an empty formatted string object
            t = FormattedString()
            # set a font
            t.font("Menlo-Regular")
            # set a font size
            t.fontSize(300)
            # add some glyphs by glyph name
            t.appendGlyph("A", "ampersand", "Eng", "Eng.alt")
            # draw the formatted string
            text(t, (100, 100))
        """
        # use a non breaking space as replacement character
        baseString = chr(0xFFFD)
        font = None
        if self._font:
            font = AppKit.NSFont.fontWithName_size_(self._font, self._fontSize)
        if font is None:
            logger.warning("font: '%s' is not installed, back to the fallback font: '%s'", self._font, _FALLBACKFONT)
            font = AppKit.NSFont.fontWithName_size_(_FALLBACKFONT, self._fontSize)

        # disable calt features, as this seems to be on by default
        # for both the font stored in the nsGlyphInfo as in the replacement character
        fontAttributes = {}
        #fontAttributes[CoreText.NSFontFeatureSettingsAttribute] = [featureMap["calt_off"]]
        fontDescriptor = font.fontDescriptor()
        fontDescriptor = fontDescriptor.fontDescriptorByAddingAttributes_(fontAttributes)
        font = AppKit.NSFont.fontWithDescriptor_size_(fontDescriptor, self._fontSize)

        fallbackFont = self._fallbackFont
        self._fallbackFont = None
        _openTypeFeatures = dict(self._openTypeFeatures)
        self._openTypeFeatures = dict(calt=False)
        for glyphName in glyphNames:
            glyph = font.glyphWithName_(glyphName)
            if glyph:
                self.append(baseString)
                glyphInfo = AppKit.NSGlyphInfo.glyphInfoWithGlyph_forFont_baseString_(glyph, font, baseString)
                self._attributedString.addAttribute_value_range_(AppKit.NSGlyphInfoAttributeName, glyphInfo, (len(self) - 1, 1))
            else:
                logger.warning("font '%s' has no glyph with the name '%s'", font.fontName(), glyphName)
        self.openTypeFeatures(**_openTypeFeatures)
        self._fallbackFont = fallbackFont



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# -----------------------------------------------------------------------------
#
#     P A G E B O T
#
#     Copyright (c) 2016+ Buro Petr van Blokland + Claudia Mens
#     www.pagebot.io
#     Licensed under MIT conditions
#
#     Supporting DrawBot, www.drawbot.com
#     Supporting Flat, xxyxyz.org/flat
# -----------------------------------------------------------------------------
#
#     tryinstallfont.py
#

import logging
from drawBot.drawBotDrawingTools import _drawBotDrawingTool

logger = logging.getLogger(__name__)

def _tryInstallFontFromFontName(fontName):
    # TODO: too much drawBot, need to port.
    try:
        return _drawBotDrawingTool._tryInstallFontFromFontName(fontName)
    except Exception as e:
        logger.error('_tryInstallFontFromFontName: %s', e)




#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
# -----------------------------------------------------------------------------
#
#     P A G E B O T
#
#     Copyright (c) 2016+ Buro Petr van Blokland + Claudia Mens
#     www.pagebot.io
#     Licensed under MIT conditions
#
#     Supporting DrawBot, www.drawbot.com
#     Supporting Flat, xxyxyz.org/flat
# -----------------------------------------------------------------------------
#
#     pattern.py
#

class FoundPattern:
    """Searches for regex patterns inside a TextLine."""

    def __init__(self, s, x, ix, y=None, w=None, h=None, line=None, run=None):
        # Found string.
        self.s = s
        self.x = x
        self.ix = ix
        self.y = y
        self.w = w
        self.h = h
        # TextLine instance that this was found in.
        self.line = line
        # List of this string.
        self.run = run

    def __repr__(self):
        return '[Found "%s" @ %d,%d]' % (self.s, self.x, self.y)

    #   F I N D

def findPattern(textLines, pattern):
    """Answers the point locations where this pattern occurs in the Formatted
    String."""
    # List of FoundPattern instances.
    foundPatterns = []

    for lineIndex, textLine in enumerate(textLines):
        for foundPattern in textLine.findPattern(pattern):
            foundPattern.y = textLine.y
            foundPattern.z = 0
            foundPatterns.append(foundPattern)
    return foundPatterns


